"""
==========================
PySpecKit Spectrum classes
==========================

The spectrum module consists of the Spectrum class, with child classes ObsBlock
and Spectra for multi-spectrum analysis of different types.

The Spectrum class is the main functional code.
ObsBlocks are containers of multiple spectra of different objects
The Spectra class is a container of multiple spectra of the *same* object at
   different wavelengths/frequencies

.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
.. moduleauthor:: Jordan Mirocha <mirochaj@gmail.com>
"""
import numpy as np
import smooth as sm
import pyfits
import readers,plotters,writers,baseline,units,measurements,speclines,arithmetic
import moments as moments_module
import fitters
import models
import history
import copy
from pyspeckit.specwarnings import warn
try:
    import atpy
    atpyOK = True
except ImportError:
    atpyOK = False

try:
    from Spectrum1D import Spectrum1D # inherit from astropy
except ImportError:
    Spectrum1D = object

class Spectrum(Spectrum1D):
    """
    The core class for the spectroscopic toolkit.  Contains the data and error
    arrays along with wavelength / frequency / velocity information in various
    formats.
    """

    from arithmetic import interpnans

    def __init__(self, filename=None, filetype=None, xarr=None, data=None,
            error=None, header=None, doplot=False, maskdata=True,
            plotkwargs={}, xarrkwargs={}, **kwargs):
        """
        Create a Spectrum object.

        Must either pass in a filename or ALL of xarr, data, and header, plus
        optionally error.

        kwargs are passed to the file reader

        Parameters
        ----------
        filename : string or pyfits.HDU
            The file to read the spectrum from.  If data, xarr, and error are
            specified, leave filename blank.
        filetype : string
            Specify the file type (only needed if it cannot be automatically
            determined from the filename)
        xarr : `units.SpectroscopicAxis` or `np.ndarray`
            The X-axis of the data.  If it is an np.ndarray, you must pass
            `xarrkwargs` or a valid header if you want to use any of the unit
            functionality.
        data : `np.ndarray`
            The data array (must have same length as xarr)
        error : `np.ndarray` 
            The error array (must have same length as the data and xarr arrays)
        header : `pyfits.Header` or dict
            The header from which to read unit information.  Needs to be a
            `pyfits.Header` instance or another dictionary-like object with the
            appropriate information
        maskdata : boolean
            turn the array into a masked array with all nan and inf values masked
        doplot : boolean
            Plot the spectrum after loading it?
        plotkwargs : dict
            keyword arguments to pass to the plotter
        xarrkwargs : dict
            keyword arguments to pass to the SpectroscopicAxis initialization
            (can be used in place of a header)

        Examples
        --------

        >>> sp = pyspeckit.Spectrum(data=np.random.randn(100),
                    xarr=np.linspace(-50, 50, 100), error=np.ones(100)*0.1, 
                    xarrkwargs={'unit':'km/s', 'refX':4.829, 'refX_units':'GHz',
                        'xtype':'VLSR-RAD'}, header={})

        >>> xarr = pyspeckit.units.SpectroscopicAxis(np.linspace(-50,50,100),
                    units='km/s', refX=6562.83, refX_units='angstroms')
        >>> data = np.random.randn(100)*5 + np.random.rand(100)*100
        >>> err = np.sqrt(data/5.)*5. # Poisson noise
        >>> sp = pyspeckit.Spectrum(data=data, error=err, xarr=xarr, header={}) 
        
        >>> # if you already have a simple fits file
        >>> sp = pyspeckit.Spectrum('test.fits')
        """

        if filename:
            if filetype is None:
                suffix = filename.rsplit('.',1)[1]
                if suffix in readers.suffix_types:
                    # use the default reader for that suffix
                    filetype = readers.suffix_types[suffix][0]
                    reader = readers.readers[filetype]
                else:
                    raise TypeError("File with suffix %s is not recognized." % suffix)
            else:
                if filetype in readers.readers:
                    reader = readers.readers[filetype]
                else:
                    raise TypeError("Filetype %s not recognized" % filetype)

            self.data,self.error,self.xarr,self.header = reader(filename,**kwargs)
            
            # these should probably be replaced with registerable function s...
            if filetype in ('fits','tspec','pyfits'):
                self.parse_header(self.header)
            elif filetype is 'txt':
                self.parse_text_header(self.header)
            elif filetype in ('hdf5', 'h5'):
                self.parse_hdf5_header(self.header)

            if isinstance(filename,str):
                self.fileprefix = filename.rsplit('.', 1)[0]    # Everything prior to .fits or .txt
        elif xarr is not None and data is not None:
            # technically, this is unpythonic.  But I don't want to search for all 10 attributes required.
            if issubclass(type(xarr),units.SpectroscopicAxis):
                self.xarr = xarr
            else:
                self.xarr = units.SpectroscopicAxis(xarr, **xarrkwargs)
            self.data = data
            if error is not None:
                self.error = error
            else:
                self.error = data * 0
            if hasattr(header,'get'):
                self.header = header
            else: # set as blank
                warn( "WARNING: Blank header." )
                self.header = pyfits.Header()
            self.parse_header(self.header)

        if maskdata:
            if hasattr(self.data,'mask'):
                self.data.mask += np.isnan(self.data) + np.isinf(self.data)
                self.error.mask += np.isnan(self.data) + np.isinf(self.data)
            else:
                self.data = np.ma.masked_where(np.isnan(self.data) + np.isinf(self.data), self.data)
                self.error = np.ma.masked_where(np.isnan(self.data) + np.isinf(self.data), self.error)

        self.plotter = plotters.Plotter(self)
        self._register_fitters()
        self.specfit = fitters.Specfit(self,Registry=self.Registry)
        self.baseline = baseline.Baseline(self)
        self.speclines = speclines
        self._sort()

        # Special.  This needs to be modified to be more flexible; for now I need it to work for nh3
        self.plot_special = None
        self.plot_special_kwargs = {}

        if doplot: self.plotter(**plotkwargs)


    def _register_fitters(self, registry=None):
        """
        Register fitters independently for each spectrum instance

        This approach allows you to add fitters to a given Spectrum instance
        without modifying the default registry
        """
        self.Registry = fitters.Registry()
        if registry is None:
            registry = fitters.default_Registry
        elif not isinstance(registry, fitters.Registry):
            raise TypeError("registry must be an instance of the fitters.Registry class")

        for modelname, model in registry.multifitters.iteritems():
            self.Registry.add_fitter(modelname, model,
                    registry.npars[modelname], multisingle='multi',
                    key=registry.associated_keys.get(modelname))
        for modelname, model in registry.singlefitters.iteritems():
            self.Registry.add_fitter(modelname, model,
                    registry.npars[modelname], multisingle='single',
                    key=registry.associated_keys.get(modelname))

    def _sort(self):
        """
        Make sure X axis is monotonic.  
        """
        if self.xarr.dxarr.min() < 0:
            argsort = np.argsort(self.xarr)
            self.data = self.data[argsort]
            self.error = self.error[argsort]
            self.xarr = self.xarr[argsort]
            self.xarr.dxarr = np.diff(self.xarr)
        
    def write(self,filename,type=None,**kwargs):
        """
        Write the spectrum to a file.  The available file types are listed
        in spectrum.writers.writers

        type - what type of file to write to?  If not specified, will attempt
        to determine type from suffix
        """
        if type:
            self.writer = writers.writers[type](self)
        else:
            suffix = filename.rsplit('.',1)[1]
            if suffix in writers.suffix_types:
                # use the default reader for that suffix
                filetype = writers.suffix_types[suffix][0]
                self.writer = writers.writers[filetype](self)
            else:
                raise TypeError("File with suffix %s is not recognized." % suffix)
        self.writer(filename=filename,**kwargs)

    def parse_text_header(self,Table):
        """
        Grab relevant parameters from a table header (xaxis type, etc)

        This function should only exist for Spectrum objects created from
        .txt or other atpy table type objects
        """
        self.Table = Table
        xtype = Table.data.dtype.names[Table.xaxcol]
        if xtype in units.xtype_dict.values():
            self.xarr.xtype = xtype
            self.xarr.units = Table.columns[xtype].unit
        elif xtype in units.xtype_dict:
            self.xarr.xtype = units.xtype_dict[xtype]
            self.xarr.units = Table.columns[xtype].unit
        else:
            warn( "Warning: Invalid xtype in text header - this may mean no text header was available.  X-axis units will be pixels unless you set them manually (e.g., sp.xarr.unit='angstroms')")
            self.xarr.xtype = 'pixels'
            self.xarr.units = 'none'
            #raise ValueError("Invalid xtype in text header")
        self.ytype = Table.data.dtype.names[Table.datacol]
        self.units = Table.columns[self.ytype].unit
        self.header = pyfits.Header()
        self._update_header()

    def _update_header(self):
        self.header.update('CUNIT1',self.xarr.units)
        self.header.update('CTYPE1',self.xarr.xtype)
        self.header.update('BUNIT',self.units)
        self.header.update('BTYPE',self.ytype)
        
    def parse_hdf5_header(self, hdr):
        """
        HDF5 reader will create a hdr dictionary from HDF5 dataset
        attributes if they exist.  This routine will convert that dict
        to a pyfits header instance.
        """    
        
        self.xarr.xtype = hdr['xtype']
        self.xarr.xunits = hdr['xunits']
        self.ytype = hdr['ytype']
        self.units = hdr['yunits']
        self.header = pyfits.Header()
        self.header.update('CUNIT1', self.xarr.xunits)
        self.header.update('CTYPE1', self.xarr.xtype)
        self.header.update('BUNIT', self.ytype)
        self.header.update('BTYPE', self.units)

    def parse_header(self,hdr,specname=None):
        """
        Parse parameters from a .fits header into required spectrum structure
        parameters

        This should be moved to the FITSSpectrum subclass when that is available
        """

        if hdr.get('BUNIT'):
            self.units = hdr.get('BUNIT').strip()
        else:
            self.units = 'undefined'
            
        if hdr.get('BTYPE'):
            self.ytype = hdr.get('BTYPE').strip()
        else:
            self.ytype = 'data'

        if specname is not None:
            self.specname = specname
        elif hdr.get('OBJECT'):
            self.specname = hdr.get('OBJECT')
        elif hdr.get('OBJNAME'):
            self.specname = hdr.get('OBJNAME')
        else:
            self.specname = ''
            
    def measure(self, z=None, d=None, fluxnorm=None, miscline=None, misctol=None, ignore=None, derive=True, **kwargs):
        """
        Initialize the measurements class - only do this after you have run a fitter otherwise pyspeckit will be angry!
        """
        self.measurements=measurements.Measurements(self, z=z, d=d,
                fluxnorm=fluxnorm, miscline=miscline, misctol=misctol,
                ignore=ignore, derive=derive, **kwargs)

    def crop(self, x1, x2, units=None, **kwargs):
        """
        Replace the current spectrum with a subset from x1 to x2 in current
        units

        Fixes CRPIX1 and baseline and model spectra to match cropped data spectrum

        """
        # do slice (this code is redundant... need to figure out how to fix that)
        x1pix = np.argmin(np.abs(x1-self.xarr.as_unit(units)))
        x2pix = np.argmin(np.abs(x2-self.xarr.as_unit(units)))
        if x1pix > x2pix: x1pix,x2pix = x2pix,x1pix
        if x1pix == x2pix:
            raise IndexError("ERROR: Trying to crop to zero size.")

        self = self.slice(x1pix, x2pix, units='pixels', copy=False, **kwargs)
        # a baseline spectrum is always defined, even if it is all zeros
        # this is needed to prevent size mismatches.  There may be a more
        # elegant way to do this...
        # this is taken care of in Slice now self.baseline.crop(x1pix,x2pix)
        # this is taken care of in Slice now self.specfit.crop(x1pix,x2pix)
        if hasattr(self.specfit, 'fitter'):
            self.specfit._full_model()

        if hasattr(self,'header'):
            history.write_history(self.header,"CROP: Cropped from %g to %g (pixel %i to %i)" % (x1,x2,x1pix,x2pix))

            if self.header.get('CRPIX1'):
                self.header.update('CRPIX1',self.header.get('CRPIX1') - x1pix)
                history.write_history(self.header,"CROP: Changed CRPIX1 from %f to %f" % (self.header.get('CRPIX1')+x1pix,self.header.get('CRPIX1')))

    def slice(self, start=None, stop=None, units='pixel', copy=True, preserve_fits=False):
        """Slicing the spectrum
        
        Parameters:
        -----------
        start : numpy.float or int
            start of slice
        stop :  numpy.float or int
            stop of slice
        units : str
            allowed values are any supported physical unit, 'pixel'
        copy : bool
            Return a 'view' of the data or a copy?
        preserve_fits : bool
            Save the fitted parameters from self.fitter?
        """
        
        if units in ('pixel','pixels'):
            start_ind = start
            stop_ind  = stop
        else:
            x_in_units = self.xarr.as_unit(units)
            start_ind = x_in_units.x_to_pix(start)
            stop_ind  = x_in_units.x_to_pix(stop)
        if start_ind > stop_ind: start_ind,stop_ind = stop_ind,start_ind
        spectrum_slice = slice(start_ind,stop_ind)

        if copy:
            sp = self.copy()
        else:
            sp=self
        sp.data = sp.data[spectrum_slice]
        if sp.error is not None:
            sp.error = sp.error[spectrum_slice]
        sp.xarr = sp.xarr[spectrum_slice]

        if copy:
            # create new specfit / baseline instances (otherwise they'll be the wrong length)
            sp._register_fitters()
            sp.baseline = baseline.Baseline(sp)
            sp.specfit = fitters.Specfit(sp,Registry=sp.Registry)
        else:
            # inplace modification
            sp.baseline.crop(start_ind, stop_ind)
            sp.specfit.crop(start_ind, stop_ind)

        if preserve_fits:
            sp.specfit.modelpars = self.specfit.modelpars
            sp.specfit.parinfo = self.specfit.parinfo
            sp.baseline.baselinepars = self.baseline.baselinepars
            sp.baseline.order = self.baseline.order


        return sp

    # For Spectrum1D compatibility, flux = data
    @property
    def flux(self):
        return self.data

    @flux.setter
    def flux(self,value):
        self.data = value
    
    def __getitem__(self, indx):
        """
        Slice the data using pixel units (not quite the same as self.slice
        because indx can be a slice object or numbers)
        """

        sp = copy.copy(self)
        sp.data = sp.data.__getitem__(indx)
        if sp.error is not None:
            sp.error = sp.error.__getitem__(indx)
        sp.xarr = sp.xarr.__getitem__(indx)

        # this should be done by deepcopy, but deepcopy fails with current pyfits
        sp.plotter = copy.copy(self.plotter)
        sp.plotter.Spectrum = sp
        sp.specfit = copy.copy(self.specfit)
        sp.specfit.Spectrum = sp
        sp.specfit.Spectrum.plotter = sp.plotter
        sp.baseline = copy.copy(self.baseline)
        sp.baseline.Spectrum = sp
        sp.baseline.Spectrum.plotter = sp.plotter
        
        return sp


    def smooth(self,smooth,**kwargs):
        """
        Smooth the spectrum by factor `smooth`.  
        """
        smooth = round(smooth)
        self.data = sm.smooth(self.data,smooth,**kwargs)
        self.xarr = self.xarr[::smooth]
        if len(self.xarr) != len(self.data):
            raise ValueError("Convolution resulted in different X and Y array lengths.  Convmode should be 'same'.")
        if self.error is not None:
            self.error = sm.smooth(self.error,smooth,**kwargs)
        self.baseline.downsample(smooth)
        self.specfit.downsample(smooth)
    
        self._smooth_header(smooth)
    smooth.__doc__ += "sm.smooth doc: \n" + sm.smooth.__doc__

    def _smooth_header(self,smooth):
        """
        Internal - correct the FITS header parameters when smoothing
        """
        if self.header.get('CDELT1') is not None and self.header.get('CRPIX1') is not None:
            self.header.update('CDELT1',self.header.get('CDELT1') * float(smooth))
            self.header.update('CRPIX1',self.header.get('CRPIX1') / float(smooth))

            history.write_history(self.header,"SMOOTH: Smoothed and downsampled spectrum by factor %i" % (smooth))
            history.write_history(self.header,"SMOOTH: Changed CRPIX1 from %f to %f" % (self.header.get('CRPIX1')*float(smooth),self.header.get('CRPIX1')))
            history.write_history(self.header,"SMOOTH: Changed CDELT1 from %f to %f" % (self.header.get('CRPIX1')/float(smooth),self.header.get('CRPIX1')))

    def _shape(self):
        """
        Return the data shape
        """
        return self.data.shape

    shape = property(_shape)

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        if hasattr(self,'specname'):
            name = " named %s" % self.specname
        else:
            name = ""
        return r'<Spectrum object%s over spectral range %6.5g : %6.5g %s and flux range = [%2.1f, %2.1f] %s at %s>' % \
                (name, self.xarr.min(), self.xarr.max(), self.xarr.units,
                        self.data.min(), self.data.max(), self.units,
                        str(hex(self.__hash__())))
    

    def copy(self,deep=True):
        """
        Create a copy of the spectrum with its own plotter, fitter, etc.
        Useful for, e.g., comparing smoothed to unsmoothed data
        """

        newspec = copy.copy(self)
        if deep:
            newspec.xarr = copy.copy(self.xarr)
            newspec.data = copy.copy(self.data)
            if self.error is not None:
                newspec.error = copy.copy(self.error)

        newspec.header = copy.copy(self.header)
        newspec.plotter = self.plotter.copy(parent=newspec)
        newspec._register_fitters()
        newspec.specfit = self.specfit.copy(parent=newspec)
        newspec.specfit.Spectrum.plotter = newspec.plotter
        newspec.baseline = self.baseline.copy(parent=newspec)
        newspec.baseline.Spectrum.plotter = newspec.plotter

        return newspec

    def stats(self, statrange=(), interactive=False):
        """
        Return some statistical measures in a dictionary (somewhat self-explanatory)

        range - X-range over which to perform measures
        interactive - specify range interactively in plotter

        """

        if len(statrange) == 2:
            pix1 = self.xarr.x_to_pix(statrange[0])
            pix2 = self.xarr.x_to_pix(statrange[1])
            if pix1 > pix2: pix1,pix2 = pix2,pix1
            elif pix1 == pix2: raise ValueError("Invalid statistics range - includes 0 pixels")
            data = self.data[pix1:pix2]
        elif interactive:
            raise NotImplementedError('Not implemented yet.  Probably need to move the stats command into a different module first')
        else:
            data = self.data

        stats = {
            "npts": data.shape[0],
            "std": data.std(),
            "mean": data.mean(),
            "median": np.median(data),
            "min": data.min(),
            "max": data.max(),}
        return stats

    def getlines(self, linetype='radio', **kwargs):
        """
        Access a registered database of spectral lines.  Will add an attribute
        with the name linetype, which then has properties defined by the
        speclines module (most likely, a table and a "show" function to display
        the lines)
        """

        # this is somewhat unreadable, but in short:
        # self.__dict__ is a dictionary that defines the class attributes
        # so, for example, if linetype is radio, this reads:
        # self.radio = speclines.radio.radio_lines(self)
        # or optical:
        # self.optical = speclines.optical.optical_lines(self)
        if not linetype in self.__dict__: # don't replace it if it already exists
            self.__dict__[linetype] = speclines.__dict__[linetype].__dict__[linetype+"_lines"](self,**kwargs)

    def moments(self, unit='km/s', **kwargs):
        """
        Return the moments of the spectrum.  In order to assure that the 1st
        and 2nd moments are meaningful, a 'default' unit is set.  If unit is not
        set, will use current unit.

        """

        if unit is False or unit is None:
            return moments_module.moments(self.xarr, self.data, **kwargs)
        else:
            return moments_module.moments(self.xarr.as_unit(unit), self.data, **kwargs)

    moments.__doc__ += moments_module.moments.__doc__

    def _operation_wrapper(operation):
        """
        Perform an operation (addition, subtraction, mutiplication, division, etc.)
        after checking for shape matching
        """

        def ofunc(self, other): 
            if np.isscalar(other):
                newspec = self.copy()
                newspec.data = operation(newspec.data, other) 
                return newspec

            elif hasattr(self,'xarr') and hasattr(other,'xarr'): # purely for readability

                if self._arithmetic_threshold == 'exact':
                    xarrcheck = all(self.xarr == other.xarr)
                else:
                    if self._arithmetic_threshold_units is None:
                        # not sure this should ever be allowed
                        xarrcheck = all((self.xarr-other.xarr) < self._arithmetic_threshold)
                    else:
                        xarrcheck = all((self.xarr.as_unit(self._arithmetic_threshold_units)-other.xarr.as_unit(self._arithmetic_threshold_units)) < self._arithmetic_threshold)

                if self.shape == other.shape and xarrcheck:
                    newspec = self.copy()
                    newspec.data = operation(newspec.data, other.data)
                    return newspec
                elif self.shape != other.shape:
                    raise ValueError("Shape mismatch in data")
                elif not xarrcheck:
                    raise ValueError("X-axes do not match.")
            elif hasattr(self,'shape') and hasattr(other,'shape'):
                # allow array subtraction
                if self.shape != other.shape:
                    raise ValueError("Shape mismatch in data")
                elif hasattr(self,'xarr'):
                    newspec = self.copy()
                    newspec.data = operation(newspec.data, other)
                elif hasattr(other,'xarr'): # is this even possible?
                    newspec = other.copy()
                    newspec.data = operation(self, other.data)

        return ofunc

    @property
    def _arithmetic_threshold(self):
        return self._arithmetic_threshold_value

    @_arithmetic_threshold.setter
    def _arithmetic_threshold(self, value, units=None):
        self._arithmetic_threshold_value = value
        if units is None:
            self._arithmetic_threshold_units = self.xarr.units
        else:
            self._arithmetic_threshold_units = units

    _arithmetic_threshold_value = 'exact'
    _arithmetic_threshold_units = None

    __add__ = _operation_wrapper(np.add)
    __radd__ = _operation_wrapper(np.add)
    __sub__ = _operation_wrapper(np.subtract)
    __mul__ = _operation_wrapper(np.multiply)
    __div__ = _operation_wrapper(np.divide)


class Spectra(Spectrum):
    """
    A list of individual Spectrum objects.  Intended to be used for
    concatenating different wavelength observations of the SAME OBJECT.  Can be
    operated on just like any Spectrum object, incuding fitting.  Useful for
    fitting multiple lines on non-continguous axes simultaneously.  Be wary of
    plotting these though...

    Can be indexed like python lists.  

    X array is forcibly sorted in increasing order
    """

    def __init__(self, speclist, xunits='GHz', **kwargs):
        print "Creating spectra"
        speclist = list(speclist)
        for ii,spec in enumerate(speclist):
            if type(spec) is str:
                spec = Spectrum(spec)
                speclist[ii] = spec

        self.speclist = speclist

        print "Concatenating data"
        self.xarr = units.SpectroscopicAxes([sp.xarr.as_unit(xunits) for sp in speclist])
        self.xarr.units = xunits 
        self.xarr.xtype = units.unit_type_dict[xunits]
        self.data = np.ma.concatenate([sp.data for sp in speclist])
        self.error = np.ma.concatenate([sp.error for sp in speclist])
        self._sort()

        self.header = pyfits.Header()
        for spec in speclist:
            for key,value in spec.header.items():
                try:
                    self.header.update(key,value)
                except ValueError, KeyError:
                    warn( "Could not update header KEY=%s to VALUE=%s" % (key,value) )

        self.plotter = plotters.Plotter(self)
        self._register_fitters()
        self.specfit = fitters.Specfit(self,Registry=self.Registry)
        self.baseline = baseline.Baseline(self)
        
        self.units = speclist[0].units
        for spec in speclist: 
            if spec.units != self.units: 
                raise ValueError("Mismatched units")

        # Special.  This needs to be modified to be more flexible; for now I need it to work for nh3
        self.plot_special = None
        self.plot_special_kwargs = {}

    def __add__(self,other):
        """
        Adding "Spectra" together will concatenate them
        * WARNING * this will probably fail now that I've implemented direct __add__...
        """
        if type(other) is Spectra or Spectrum:
            self.speclist += other.speclist
        elif type(other) is Spectrum:
            self.speclist += [other.speclist]
        if other.units != self.units:
            raise ValueError("Mismatched units")

        if other.xarr.units != self.xarr.units:
            # convert all inputs to same units
            spec.xarr.convert_to_units(self.xarr.units,**kwargs)
        self.xarr = units.SpectroscopicAxes([self.xarr,spec.xarr])
        self.data = np.concatenate([self.data,spec.data])
        self.error = np.concatenate([self.error,spec.error])
        self._sort()

    def __getitem__(self,index):
        """
        Can index Spectra to get the component Spectrum objects
        """
        return self.speclist[index]

    def __len__(self): 
        """
        len(spectra) != len(spectrum) !
        """
        return len(self.speclist)

    def smooth(self,smooth,**kwargs):
        """
        Smooth the spectrum by factor "smooth".  Options are defined in sm.smooth

        because 'Spectra' does not have a header attribute, don't do anything to it...
        """
        smooth = round(smooth)
        self.data = sm.smooth(self.data,smooth,**kwargs)
        self.xarr = self.xarr[::smooth]
        if len(self.xarr) != len(self.data):
            raise ValueError("Convolution resulted in different X and Y array lengths.  Convmode should be 'same'.")
        self.error = sm.smooth(self.error,smooth,**kwargs)
        self.baseline.downsample(smooth)
        self.specfit.downsample(smooth)

    def fiteach(self,**kwargs):
        """
        Fit each spectrum within the Spectra object
        """
        for sp in self.speclist:
            sp.specfit(**kwargs)

        if atpyOK:
            self.fittable = atpy.Table()
            self.fittable.add_column('name',[sp.specname for sp in self.speclist])
            self.fittable.add_column('amplitude',[sp.specfit.modelpars[0] for sp in self.speclist],unit=self.units)
            self.fittable.add_column('center',[sp.specfit.modelpars[1] for sp in self.speclist],unit=self.xarr.units)
            self.fittable.add_column('width',[sp.specfit.modelpars[2] for sp in self.speclist],unit=self.xarr.units)
            self.fittable.add_column('amplitudeerr',[sp.specfit.modelerrs[0] for sp in self.speclist],unit=self.units)
            self.fittable.add_column('centererr',[sp.specfit.modelerrs[1] for sp in self.speclist],unit=self.xarr.units)
            self.fittable.add_column('widtherr',[sp.specfit.modelerrs[2] for sp in self.speclist],unit=self.xarr.units)

    def ploteach(self, xunits=None, inherit_fit=False, plot_fit=True, plotfitkwargs={}, **plotkwargs):
        """
        Plot each spectrum in its own window
        inherit_fit - if specified, will grab the fitter & fitter properties from Spectra
        """
        for sp in self.speclist:
            if xunits is not None:
                sp.xarr.convert_to_unit(xunits,quiet=True)
            if inherit_fit:
                sp.specfit.fitter = self.specfit.fitter
                sp.specfit.modelpars = self.specfit.modelpars
                sp.specfit.model = np.interp(sp.xarr.as_unit(self.xarr.units),self.xarr,self.specfit.fullmodel)

            sp.plotter(**plotkwargs)
            
            if plot_fit and self.specfit.model is not None:
                sp.specfit.plot_fit(**plotfitkwargs)


class ObsBlock(Spectra):
    """
    An Observation Block

    Consists of multiple spectra with a shared X-axis.  Intended to hold groups
    of observations of the same object in the same setup for later averaging.

    ObsBlocks can be indexed like python lists.  
    """

    def __init__(self, speclist, xtype='frequency', xarr=None, force=False, **kwargs):

        if xarr is None:
            self.xarr = speclist[0].xarr
        else:
            self.xarr = xarr

        self.units = speclist[0].units
        self.header = speclist[0].header
        self.parse_header(self.header)

        for spec in speclist:
            if not isinstance(spec,Spectrum):
                raise TypeError("Must create an ObsBlock with a list of spectra.")
            if not np.array_equal(spec.xarr, self.xarr):
                if not force:
                    raise ValueError("Mismatch between X axes in ObsBlock")
            if spec.units != self.units: 
                raise ValueError("Mismatched units")

        if force:
            self.speclist = [arithmetic.interp(spec,self) for spec in speclist]
        else:
            self.speclist = speclist
        self.nobs = len(self.speclist)

        # Create a 2-dimensional array of the data
        self.data = np.array([sp.data for sp in self.speclist]).swapaxes(0,1).squeeze()
        self.error = np.array([sp.error for sp in self.speclist]).swapaxes(0,1).squeeze()

        self.plotter = plotters.Plotter(self)
        self._register_fitters()
        self.specfit = fitters.Specfit(self,Registry=self.Registry)
        self.baseline = baseline.Baseline(self)
        
    def average(self, weight=None, inverse_weight=False, error='erravgrtn', debug=False):
        """
        Average all scans in an ObsBlock.  Returns a single Spectrum object
        
        Parameters
        ----------
        weight : string
            a header keyword to weight by.   If not specified, the spectra will be 
            averaged without weighting
        inverse_weight : bool
            Is the header keyword an inverse-weight (e.g., a variance?)
        error : ['scanrms','erravg','erravgrtn']
            estimate the error spectrum by one of three methods.
            'scanrms'   : the standard deviation of each pixel across all scans
            'erravg'    : the average of all input error spectra
            'erravgrtn' : the average of all input error spectra divided by sqrt(n_obs)
        """

        wtarr = np.isfinite(self.data)
        if weight is not None:
            if inverse_weight:
                for ii,sp in enumerate(self.speclist):
                    wtarr[:,ii] *= 1.0/sp.header.get(weight)
            else:
                for ii,sp in enumerate(self.speclist):
                    wtarr[:,ii] *= sp.header.get(weight)

        if self.header.get('EXPOSURE'):
            self.header['EXPOSURE'] = np.sum([sp.header['EXPOSURE'] for sp in self.speclist])

        data_nonan = np.nan_to_num(self.data)
        weighted_data = (data_nonan * wtarr)
        weighted_data_axsum = weighted_data.sum(axis=1)
        weight_axsum = wtarr.sum(axis=1) 
        avgdata =  weighted_data_axsum / weight_axsum 
        if error is 'scanrms':
            # axis swapping is for projection... avgdata = 0'th axis
            errspec = np.sqrt( (((data_nonan.swapaxes(0,1)-avgdata) * wtarr.swapaxes(0,1))**2 / wtarr.swapaxes(0,1)**2).swapaxes(0,1).sum(axis=1) )
        elif error is 'erravg':
            errspec = self.error.mean(axis=1)
        elif error is 'erravgrtn':
            errspec = self.error.mean(axis=1) / np.sqrt(self.error.shape[1])

        spec = Spectrum(data=avgdata,error=errspec,xarr=self.xarr.copy(),header=self.header)
        spec._arithmetic_threshold = self._arithmetic_threshold

        if debug:
            # this statement, and much of the text above, is to test an absolutely insane error:
            # wtarr.sum(axis=1) randomly - say, one out of every 10-100 occurrences - fills with 
            # nonsense values (1e-20, 1e-55, whatever).  There is no pattern to this; it occurs in
            # while loops, but ONLY IN THIS FUNCTION.  This is unreproduceable anywhere else.
            print "selfdata    min: %10g max: %10g" % (self.data.min(), self.data.max())
            print "nonandata   min: %10g max: %10g" % (data_nonan.min(), data_nonan.max())
            print "avgdata     min: %10g max: %10g" % (avgdata.min(), avgdata.max())
            print "weight      sum: %10g" % (wtarr.sum())
            print "data*weight sum: %10g" % ((data_nonan*wtarr).sum())
            if np.abs(data_nonan.min()/avgdata.min()) > 1e10:
                import pdb; pdb.set_trace()

        return spec

    def __len__(self): return len(self.speclist)

    def __getitem__(self,index):
        """
        Can index Spectra to get the component Spectrum objects
        """
        return self.speclist[index]

    def smooth(self,smooth,**kwargs):
        """
        Smooth the spectrum by factor "smooth".  Options are defined in sm.smooth
        """
        smooth = round(smooth)
        self.data = sm.smooth_multispec(self.data,smooth,**kwargs)
        self.xarr = self.xarr[::smooth]
        if len(self.xarr) != len(self.data):
            raise ValueError("Convolution resulted in different X and Y array lengths.  Convmode should be 'same'.")
        self.error = sm.smooth_multispec(self.error,smooth,**kwargs)
        self.baseline.downsample(smooth)
        self.specfit.downsample(smooth)
    
        self._smooth_header(smooth)

class XCorrSpectrum(Spectrum):
    """ extraordinarily thin spectrum; just a name right now """
    pass

if __name__ == "__main__":
    import doctest
    doctest.testmod()




from numpy import sign

def ratos(ra):
    r = ra/15.
    deg = r - (r % 1)
    deg -= (deg % 1)
    min = (r % 1) * 60.
    sec = (min % 1) * 60
    min -= (min % 1)
    rs = "%02.0f:%02.0f:%05.3f" % (deg,min,sec)
    return rs

def dectos(dec):
    deg = sign(dec) * (abs(dec) - (abs(dec) % 1))
    deg -= (deg%1)
    min = (abs(dec) % 1) * 60.
    sec = (min % 1) * 60
    min -= (min % 1)
    decs = "%+02.0f:%02.0f:%05.3f" % (deg,min,sec)
    return decs




writers = {}
suffix_types = {}

class Writer(object):
    def __init__(self, Spectrum):
        self.Spectrum = Spectrum
        
    def __call__(self,**kwargs):
        self.write_data(**kwargs)

    def write_data(self):
        print "This is a dummy function intended to be overloaded."

import hdf5_writer
write_hdf5 = hdf5_writer.write_hdf5
import fits_writer
write_fits = fits_writer.write_fits



"""
===========================
Formaldehyde cm-line fitter
===========================

This is a formaldehyde 1_11-1_10 / 2_12-2_11 fitter.  It includes hyperfine
components of the formaldehyde lines and has both LTE and RADEX LVG based
models
"""
import numpy as np
from pyspeckit.mpfit import mpfit
from .. import units
from . import fitter,model,modelgrid
import matplotlib.cbook as mpcb
import copy
import hyperfine
from pyspeckit.specwarnings import warn
try: # for model grid reading
    import pyfits
    pyfitsOK = True
except:
    pyfitsOK = False
try:
    import scipy.interpolate
    import scipy.ndimage
    scipyOK = True
except ImportError:
    scipyOK=False

line_names = ['oneone','twotwo','threethree']
line_names = ['oneone_f10','oneone_f01','oneone_f22','oneone_f21','oneone_f12','oneone_f11','twotwo_f11','twotwo_f12','twotwo_f21','twotwo_f32','twotwo_f33','twotwo_f22','twotwo_f23']

ckms = units.speedoflight_ms / 1e3 #2.99792458e5

# http://adsabs.harvard.edu/abs/1971ApJ...169..429T has the most accurate freqs
# http://adsabs.harvard.edu/abs/1972ApJ...174..463T [twotwo]
central_freq_dict = { 
    'oneone':     4.82965996e9,
    'twotwo':     14.48847881e9,
    'threethree': 28.97480e9,
    }
line_strength_dict={
        'oneone_f10':  4.,
        'oneone_f01':  4.,
        'oneone_f22': 15.,
        'oneone_f21':  5.,
        'oneone_f12':  5.,
        'oneone_f11':  3.,
        'twotwo_f11': 15.,
        'twotwo_f12':  5.,
        'twotwo_f21':  5.,
        'twotwo_f32': 5.19,
        'twotwo_f33': 41.48,
        'twotwo_f22': 23.15,
        'twotwo_f23': 5.19,
        'threethree_f22':1,
        'threethree_f44':1,
        'threethree_f33':1,
        }
relative_strength_total_degeneracy={
        'oneone_f10': 36.,
        'oneone_f01': 36.,
        'oneone_f22': 36.,
        'oneone_f21': 36.,
        'oneone_f12': 36.,
        'oneone_f11': 36.,
        'twotwo_f11': 100.01,
        'twotwo_f12': 100.01,
        'twotwo_f21': 100.01,
        'twotwo_f32': 100.01,
        'twotwo_f33': 100.01,
        'twotwo_f22': 100.01,
        'twotwo_f23': 100.01,
        'threethree_f22':3.0,
        'threethree_f44':3.0,
        'threethree_f33':3.0,
        }
hf_freq_dict={
        'oneone_f10':4.82965996e9 - 18.53e3,
        'oneone_f01':4.82965996e9 - 1.34e3,
        'oneone_f22':4.82965996e9 - 0.35e3,
        'oneone_f21':4.82965996e9 + 4.05e3,
        'oneone_f12':4.82965996e9 + 6.48e3,
        'oneone_f11':4.82965996e9 + 11.08e3,
        'twotwo_f11':14.48847881e9 - 19.97e3,
        'twotwo_f12':14.48847881e9 -  7.03e3,
        'twotwo_f21':14.48847881e9 -  2.20e3,
        'twotwo_f32':14.48847881e9 +  0.12e3,
        'twotwo_f33':14.48847881e9 +  0.89e3,
        'twotwo_f22':14.48847881e9 + 10.74e3,
        'twotwo_f23':14.48847881e9 + 11.51e3,
        'threethree_f22':28.97478e9,
        'threethree_f44':28.97480e9,
        'threethree_f33':28.97481e9,
        }
freq_dict = dict(hf_freq_dict.items() + central_freq_dict.items())
aval_dict = {
    'oneone':     10**-8.44801,  #64*!pi**4/(3*h*c**3)*nu11**3*mu0**2*(1/2.)
    'twotwo':     10**-7.49373,  #64*!pi**4/(3*h*c**3)*nu22**3*mu0**2*(2/3.)
    'threethree': 10**-6.89179,  #64*!pi**4/(3*h*c**3)*nu33**3*mu0**2*(3/4.)
    }
hf_aval_dict={
        'oneone_f10':10**-8.92509,
        'oneone_f01':10**-8.44797,
        'oneone_f22':10**-8.57294,
        'oneone_f21':10**-9.05004,
        'oneone_f12':10**-8.82819,
        'oneone_f11':10**-9.05009,
        'twotwo_f11':10**-7.61876,
        'twotwo_f12':10**-8.09586,
        'twotwo_f21':10**-8.31771,
        'twotwo_f32':10**-8.44804,
        'twotwo_f33':10**-7.54494,
        'twotwo_f22':10**-7.65221,
        'twotwo_f23':10**-8.30191,
        'threethree_f22':10**-6.94294,
        'threethree_f44':10**-6.91981,
        'threethree_f33':10**-6.96736,
        }
ortho_dict = {
    'oneone':     False,
    'twotwo':     False,
    'threethree': False,
    }
n_ortho = np.arange(0,28,3) # 0..3..27
n_para = np.array([x for x in range(28) if x % 3 != 0])

voff_lines_dict = {
        'oneone': [(hf_freq_dict[f]-freq_dict['oneone'])/freq_dict['oneone']*units.speedoflight_ms for f in hf_freq_dict.keys() if "oneone" in f],
        'twotwo': [(hf_freq_dict[f]-freq_dict['twotwo'])/freq_dict['twotwo']*units.speedoflight_ms for f in hf_freq_dict.keys() if "twotwo" in f],
        'threethree': [(hf_freq_dict[f]-freq_dict['threethree'])/freq_dict['threethree']*units.speedoflight_ms for f in hf_freq_dict.keys() if "threethree" in f],
        }
voff_lines_dict={ # opposite signs of freq offset
        'oneone_f10': + 18.53e3/freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'oneone_f01': + 1.34e3 /freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'oneone_f22': + 0.35e3 /freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'oneone_f21': - 4.05e3 /freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'oneone_f12': - 6.48e3 /freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'oneone_f11': - 11.08e3/freq_dict['oneone'] * units.speedoflight_ms / 1000.0,
        'twotwo_f11': + 19.97e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f12': +  7.03e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f21': +  2.20e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f32': -  0.12e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f33': -  0.89e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f22': - 10.74e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'twotwo_f23': - 11.51e3/freq_dict['twotwo'] * units.speedoflight_ms / 1000.0,
        'threethree_f22':28.97478e9,
        'threethree_f44':28.97480e9,
        'threethree_f33':28.97481e9,
        }


formaldehyde_vtau = hyperfine.hyperfinemodel(line_names, voff_lines_dict, freq_dict, line_strength_dict, relative_strength_total_degeneracy)
formaldehyde_vtau_fitter = formaldehyde_vtau.fitter
formaldehyde_vtau_vheight_fitter = formaldehyde_vtau.vheight_fitter

def formaldehyde_radex(xarr, density=4, column=13, xoff_v=0.0, width=1.0, 
        grid_vwidth=1.0,
        grid_vwidth_scale=False,
        texgrid=None,
        taugrid=None,
        hdr=None,
        path_to_texgrid='',
        path_to_taugrid='',
        temperature_gridnumber=3,
        debug=False,
        verbose=False,
        **kwargs):
    """
    Use a grid of RADEX-computed models to make a model line spectrum

    The RADEX models have to be available somewhere.
    OR they can be passed as arrays.  If as arrays, the form should be:
    texgrid = ((minfreq1,maxfreq1,texgrid1),(minfreq2,maxfreq2,texgrid2))

    xarr must be a SpectroscopicAxis instance
    xoff_v, width are both in km/s

    grid_vwidth is the velocity assumed when computing the grid in km/s
        this is important because tau = modeltau / width (see, e.g., 
        Draine 2011 textbook pgs 219-230)
    grid_vwidth_scale is True or False: False for LVG, True for Sphere
    """

    if texgrid is None and taugrid is None:
        if path_to_texgrid == '' or path_to_taugrid=='':
            raise IOError("Must specify model grids to use.")
        else:
            taugrid = [pyfits.getdata(path_to_taugrid)]
            texgrid = [pyfits.getdata(path_to_texgrid)]
            hdr = pyfits.getheader(path_to_taugrid)
            yinds,xinds = np.indices(taugrid[0].shape[1:])
            densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
            columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
            minfreq = (4.8,)
            maxfreq = (5.0,)
    elif len(taugrid)==len(texgrid) and hdr is not None:
        minfreq,maxfreq,texgrid = zip(*texgrid)
        minfreq,maxfreq,taugrid = zip(*taugrid)
        yinds,xinds = np.indices(taugrid[0].shape[1:])
        densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
        columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
    else:
        raise Exception
    
    # Convert X-units to frequency in GHz
    xarr = xarr.as_unit('Hz', quiet=True)

    tau_nu_cumul = np.zeros(len(xarr))

    gridval1 = np.interp(density, densityarr[0,:], xinds[0,:])
    gridval2 = np.interp(column, columnarr[:,0], yinds[:,0])
    if np.isnan(gridval1) or np.isnan(gridval2):
        raise ValueError("Invalid column/density")

    if scipyOK:
        slices = [temperature_gridnumber] + [slice(np.floor(gv),np.floor(gv)+2) for gv in (gridval2,gridval1)]
        tau = [scipy.ndimage.map_coordinates(tg[slices],np.array([[gridval2%1],[gridval1%1]]),order=1) for tg in taugrid]
        tex = [scipy.ndimage.map_coordinates(tg[slices],np.array([[gridval2%1],[gridval1%1]]),order=1) for tg in texgrid]
    else:
        raise ImportError("Couldn't import scipy, therefore cannot interpolate")
    #tau = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,taugrid[temperature_gridnumber,:,:])
    #tex = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,texgrid[temperature_gridnumber,:,:])

    if verbose:
        print "density %20.12g column %20.12g: tau %20.12g tex %20.12g" % (density, column, tau, tex)

    if debug:
        import pdb; pdb.set_trace()

    spec = np.sum([(formaldehyde_vtau(xarr,Tex=float(tex[ii]),tau=float(tau[ii]),xoff_v=xoff_v,width=width, **kwargs)
                * (xarr.as_unit('GHz')>minfreq[ii]) * (xarr.as_unit('GHz')<maxfreq[ii])) for ii in xrange(len(tex))],
                axis=0)
  
    return spec

def formaldehyde_radex_orthopara_temp(xarr, density=4, column=13, 
        orthopara=1.0,
        temperature=15.0,
        xoff_v=0.0, width=1.0, 
        grid_vwidth=1.0,
        grid_vwidth_scale=False,
        texgrid=None,
        taugrid=None,
        hdr=None,
        path_to_texgrid='',
        path_to_taugrid='',
        debug=False,
        verbose=False,
        **kwargs):
    """
    Use a grid of RADEX-computed models to make a model line spectrum

    The RADEX models have to be available somewhere.
    OR they can be passed as arrays.  If as arrays, the form should be:
    texgrid = ((minfreq1,maxfreq1,texgrid1),(minfreq2,maxfreq2,texgrid2))

    xarr must be a SpectroscopicAxis instance
    xoff_v, width are both in km/s

    grid_vwidth is the velocity assumed when computing the grid in km/s
        this is important because tau = modeltau / width (see, e.g., 
        Draine 2011 textbook pgs 219-230)
    grid_vwidth_scale is True or False: False for LVG, True for Sphere
    """

    if texgrid is None and taugrid is None:
        if path_to_texgrid == '' or path_to_taugrid=='':
            raise IOError("Must specify model grids to use.")
        else:
            taugrid = [pyfits.getdata(path_to_taugrid)]
            texgrid = [pyfits.getdata(path_to_texgrid)]
            hdr = pyfits.getheader(path_to_taugrid)
            minfreq = (4.8,)
            maxfreq = (5.0,)
    elif len(taugrid)==len(texgrid) and hdr is not None:
        minfreq,maxfreq,texgrid = zip(*texgrid)
        minfreq,maxfreq,taugrid = zip(*taugrid)
    else:
        raise Exception

    densityarr = (np.arange(taugrid[0].shape[3])+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
    columnarr  = (np.arange(taugrid[0].shape[2])+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
    temparr  = (np.arange(taugrid[0].shape[1])+hdr['CRPIX3']-1)*hdr['CDELT3']+hdr['CRVAL3'] # temperature
    oprarr  = (np.arange(taugrid[0].shape[0])+hdr['CRPIX4']-1)*hdr['CDELT4']+hdr['CRVAL4'] # log ortho/para ratio
    
    tau_nu_cumul = np.zeros(len(xarr))

    gridval1 = np.interp(density,     densityarr,  np.arange(len(densityarr)))
    gridval2 = np.interp(column,      columnarr,   np.arange(len(columnarr)))
    gridval3 = np.interp(temperature, temparr,     np.arange(len(temparr)))
    gridval4 = np.interp(orthopara,   oprarr,      np.arange(len(oprarr)))
    if np.isnan(gridval1) or np.isnan(gridval2):
        raise ValueError("Invalid column/density")

    if scipyOK:
        slices = [slice(int(np.floor(gv)),int(np.floor(gv)+2)) for gv in (gridval4,gridval3,gridval2,gridval1)]
        tau = [scipy.ndimage.map_coordinates(tg[slices],np.array([[gridval4%1],[gridval3%1],[gridval2%1],[gridval1%1]]),order=1,prefilter=False) for tg in taugrid]
        tex = [scipy.ndimage.map_coordinates(tg[slices],np.array([[gridval4%1],[gridval3%1],[gridval2%1],[gridval1%1]]),order=1,prefilter=False) for tg in texgrid]
    else:
        raise ImportError("Couldn't import scipy, therefore cannot interpolate")
    #tau = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,taugrid[temperature_gridnumber,:,:])
    #tex = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,texgrid[temperature_gridnumber,:,:])

    if verbose:
        print "density %20.12g   column: %20.12g   temperature: %20.12g   opr: %20.12g   xoff_v: %20.12g   width: %20.12g" % (density, column, temperature, orthopara, xoff_v, width)
        print "tau: ",tau," tex: ",tex
        print "minfreq: ",minfreq," maxfreq: ",maxfreq

    if debug > 1:
        import pdb; pdb.set_trace()

    spec = np.sum([(formaldehyde_vtau(xarr.as_unit('Hz',quiet=True),Tex=float(tex[ii]),tau=float(tau[ii]),xoff_v=xoff_v,width=width, **kwargs)
                * (xarr.as_unit('GHz')>minfreq[ii]) * (xarr.as_unit('GHz')<maxfreq[ii])) for ii in xrange(len(tex))],
                axis=0)
  
    return spec


def formaldehyde(xarr, amp=1.0, xoff_v=0.0, width=1.0, 
        return_hyperfine_components=False ):
    """
    Generate a model Formaldehyde spectrum based on simple gaussian parameters

    the "amplitude" is an essentially arbitrary parameter; we therefore define it to be Tex given tau=0.01 when
    passing to the fitter
    The final spectrum is then rescaled to that value
    """

    mdl = formaldehyde_vtau(xarr, Tex=amp*0.01, tau=0.01, xoff_v=xoff_v, width=width, return_hyperfine_components=return_hyperfine_components)
    if return_hyperfine_components:
        mdlpeak = np.abs(mdl).squeeze().sum(axis=0).max()
    else:
        mdlpeak = np.abs(mdl).max()
    if mdlpeak > 0:
        mdl *= amp/mdlpeak

    return mdl


class formaldehyde_model(model.SpectralModel):
    def formaldehyde_integral(self, modelpars, linename='oneone'):
        """
        Return the integral of the individual components (ignoring height)
        """
        # produced by directly computing the integral of gaussians and formaldehydeians as a function of 
        # line width and then fitting that with a broken logarithmic power law
        # The errors are <0.5% for all widths
        formaldehyde_to_gaussian_ratio_coefs = {
                'lt0.1_oneone': np.array([ -5.784020,-40.058798,-111.172706,-154.256411,-106.593122,-28.933119]),
                'gt0.1_oneone': np.array([  0.038548, -0.071162, -0.045710,  0.183828, -0.145429,  0.040039]),
                'lt0.1_twotwo': np.array([  1.156561,  6.638570, 11.782065, -0.429536,-24.860297,-27.902274, -9.510288]),
                'gt0.1_twotwo': np.array([ -0.090646,  0.078204,  0.123181, -0.175590,  0.089506, -0.034687,  0.008676]),
                }


        integ = 0
        if len(modelpars) % 3 == 0:
            for amp,cen,width in np.reshape(modelpars,[len(modelpars)/3,3]):
                gaussint = amp*width*np.sqrt(2.0*np.pi)
                cftype = "gt0.1_"+linename if width > 0.1 else "lt0.1_"+linename
                correction_factor = 10**np.polyval(formaldehyde_to_gaussian_ratio_coefs[cftype], np.log10(width) )
                # debug statement print "Two components of the integral: amp %g, width %g, gaussint %g, correction_factor %g " % (amp,width,gaussint,correction_factor)
                integ += gaussint*correction_factor

        return integ

formaldehyde_fitter = formaldehyde_model(formaldehyde, 3,
        parnames=['amp','center','width'], 
        parlimited=[(False,False),(False,False), (True,False)], 
        parlimits=[(0,0), (0,0), (0,0)],
        shortvarnames=("A","v","\\sigma"), # specify the parameter names (TeX is OK)
        fitunits='Hz' )

formaldehyde_vheight_fitter = formaldehyde_model(fitter.vheightmodel(formaldehyde), 4,
        parnames=['height','amp','center','width'], 
        parlimited=[(False,False),(False,False),(False,False), (True,False)], 
        parlimits=[(0,0), (0,0), (0,0), (0,0)],
        shortvarnames=("H","A","v","\\sigma"), # specify the parameter names (TeX is OK)
        fitunits='Hz' )


try:
    import pymodelfit

    class pmfFormaldehydeModel(pymodelfit.FunctionModel1DAuto):
        def f(self, x, amp0=1.0, xoff_v0=0.0,width0=1.0):
            return formaldehyde(x,
                    amp=amp0,
                    xoff_v=xoff_v0,width=width0)

    class pmfFormaldehydeModelVtau(pymodelfit.FunctionModel1DAuto):
        def f(self, x, Tex0=1.0, tau0=0.01, xoff_v0=0.0, width0=1.0):
            return formaldehyde_vtau(x,
                    Tex=Tex0, tau=tau0,
                    xoff_v=xoff_v0,width=width0)
except ImportError:
    pass



"""
===========
N2H+ fitter
===========
Reference for line params: 
Daniel, F., Dubernet, M.-L., Meuwly, M., Cernicharo, J., Pagani, L. 2005, MNRAS 363, 1083

http://www.strw.leidenuniv.nl/~moldata/N2H+.html

http://adsabs.harvard.edu/abs/2005MNRAS.363.1083D

Does not yet implement:
    http://adsabs.harvard.edu/abs/2010ApJ...716.1315K

"""
import numpy as np
from pyspeckit.mpfit import mpfit
from .. import units
from . import fitter,model,modelgrid
import matplotlib.cbook as mpcb
import copy
try: # for model grid reading
    import pyfits
    pyfitsOK = True
except:
    pyfitsOK = False
try:
    import scipy.interpolate
    import scipy.ndimage
    scipyOK = True
except ImportError:
    scipyOK=False
import hyperfine


freq_dict={
'110-011':93.171617e9,
'112-011':93.171913e9,
'112-012':93.171913e9,
'111-010':93.172048e9,
'111-011':93.172048e9,
'111-012':93.172048e9,
'122-011':93.173475e9,
'122-012':93.173475e9,
'123-012':93.173772e9,
'121-010':93.173963e9,
'121-011':93.173963e9,
'121-012':93.173963e9,
'101-010':93.176261e9,
'101-011':93.176261e9,
'101-012':93.176261e9,
}
aval_dict = {
'110-011':3.628,
'112-011':0.907,
'112-012':2.721,
'111-010':1.209,
'111-011':0.907,
'111-012':1.512,
'122-011':2.721,
'122-012':0.907,
'123-012':3.628,
'121-010':2.015,
'121-011':1.512,
'121-012':0.101,
'101-010':0.403,
'101-011':1.209,
'101-012':2.016,
}

line_strength_dict = { # effectively the degeneracy per rotation state...
'110-011':0.333,
'112-011':0.417,
'112-012':1.250,
'111-010':0.333,
'111-011':0.250,
'111-012':0.417,
'122-011':1.250,
'122-012':0.417,
'123-012':2.330,
'121-010':0.556,
'121-011':0.417,
'121-012':0.028,
'101-010':0.111,
'101-011':0.333,
'101-012':0.55,
}
relative_strength_total_degeneracy = {
'110-011':9.0,
'112-011':9.0,
'112-012':9.0,
'111-010':9.0,
'111-011':9.0,
'111-012':9.0,
'122-011':9.0,
'122-012':9.0,
'123-012':9.0,
'121-010':9.0,
'121-011':9.0,
'121-012':9.0,
'101-010':9.0,
'101-011':9.0,
'101-012':9.0,
}

"""
Line strengths of the 15 hyperfine components in J=1-0  transition. The
thickness of the lines indicates their relative weight compared to the others.
Line strengths are normalized in such a way that summing over all initial J = 1
levels gives the degeneracy of the J = 0 levels, i.e.,  for JF1F  012,
three for JF1F  011, and one for JF1F  010. Thus, the sum over all 15
transitions gives the total spin degeneracy
"""

line_names = freq_dict.keys()

ckms = units.speedoflight_ms / 1e3 #2.99792458e5
voff_lines_dict = dict([(k,(v-93.176261e9)/93.176261e9*ckms) for k,v in freq_dict.iteritems()])

n2hp_vtau = hyperfine.hyperfinemodel(line_names, voff_lines_dict, freq_dict, line_strength_dict, relative_strength_total_degeneracy)
n2hp_vtau_fitter = n2hp_vtau.fitter
n2hp_vtau_vheight_fitter = n2hp_vtau.vheight_fitter

def n2hp_radex(xarr, density=4, column=13, xoff_v=0.0, width=1.0, 
        grid_vwidth=1.0,
        grid_vwidth_scale=False,
        texgrid=None,
        taugrid=None,
        hdr=None,
        path_to_texgrid='',
        path_to_taugrid='',
        temperature_gridnumber=3,
        debug=False,
        verbose=False,
        **kwargs):
    """
    Use a grid of RADEX-computed models to make a model line spectrum

    The RADEX models have to be available somewhere.
    OR they can be passed as arrays.  If as arrays, the form should be:
    texgrid = ((minfreq1,maxfreq1,texgrid1),(minfreq2,maxfreq2,texgrid2))

    xarr must be a SpectroscopicAxis instance
    xoff_v, width are both in km/s

    grid_vwidth is the velocity assumed when computing the grid in km/s
        this is important because tau = modeltau / width (see, e.g., 
        Draine 2011 textbook pgs 219-230)
    grid_vwidth_scale is True or False: False for LVG, True for Sphere
    """

    if texgrid is None and taugrid is None:
        if path_to_texgrid == '' or path_to_taugrid=='':
            raise IOError("Must specify model grids to use.")
        else:
            taugrid = [pyfits.getdata(path_to_taugrid)]
            texgrid = [pyfits.getdata(path_to_texgrid)]
            hdr = pyfits.getheader(path_to_taugrid)
            yinds,xinds = np.indices(taugrid[0].shape[1:])
            densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
            columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
            minfreq = (4.8,)
            maxfreq = (5.0,)
    elif len(taugrid)==len(texgrid) and hdr is not None:
        minfreq,maxfreq,texgrid = zip(*texgrid)
        minfreq,maxfreq,taugrid = zip(*taugrid)
        yinds,xinds = np.indices(taugrid[0].shape[1:])
        densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
        columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
    else:
        raise Exception
    
    # Convert X-units to frequency in GHz
    xarr = copy.copy(xarr)
    xarr.convert_to_unit('Hz', quiet=True)

    tau_nu_cumul = np.zeros(len(xarr))

    gridval1 = np.interp(density, densityarr[0,:], xinds[0,:])
    gridval2 = np.interp(column, columnarr[:,0], yinds[:,0])
    if np.isnan(gridval1) or np.isnan(gridval2):
        raise ValueError("Invalid column/density")

    if scipyOK:
        tau = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[gridval2],[gridval1]]),order=1) for tg in taugrid]
        tex = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[gridval2],[gridval1]]),order=1) for tg in texgrid]
    else:
        raise ImportError("Couldn't import scipy, therefore cannot interpolate")
    #tau = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,taugrid[temperature_gridnumber,:,:])
    #tex = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,texgrid[temperature_gridnumber,:,:])

    if verbose:
        print "density %20.12g column %20.12g: tau %20.12g tex %20.12g" % (density, column, tau, tex)

    if debug:
        import pdb; pdb.set_trace()

    return n2hp_vtau(xarr,Tex=tex,tau=tau,xoff_v=xoff_v,width=width,**kwargs)




try:
    import scipy
    from scipy import optimize,sqrt
    from scipy.optimize import leastsq
    #from scipy.stats.stats import nanmedian,nanmean,_nanmedian
except ImportError:
    print "Scipy cold not be loaded.  Collapse_gaussfit may fail"
import numpy
from numpy import vectorize,zeros,exp,median,where,asarray,array,nonzero,ma,arange,square
import matplotlib
#matplotlib.use('Agg')
from pylab import indices,figure,clf,savefig,plot,legend,text,axes,title
import pickle
import pyfits
import time
from mad import MAD
from ratosexagesimal import ratos,dectos

def nanmedian(arr):
    """ nanmedian - this version is NOT capable of broadcasting (operating along axes) """
    return median(arr[arr==arr])
def nanmean(arr):
    """ nanmean - this version is NOT capable of broadcasting (operating along axes) """
    return (arr[arr==arr]).mean()

# read in file
# filename = sys.argv[1]
# fitsfile = pyfits.open(filename)
# cube = fitsfile[0].data

# def gaussian(dx,sigma):
#     return lambda x: exp( - (x-dx)**2 / sigma**2 )
# def return_param(xarr,param):
#     errorfunction = lambda p:gaussian(*p)(*indices(xarr.shape))-xarr
#     pars, cov, infodict, errmsg, success = optimize.leastsq(errorfunction, [len(xarr)/2.,1], full_output=1)
#     print errmsg
#     if param == 'width':
#         return pars[1]
#     elif param == 'center':
#         return pars[0]
#     else:
#         return 
def gaussian(dx,sigma,a):
    return lambda x: a*exp( - (x-dx)**2 / sigma**2 )
def double_gaussian(dx1,dx2,sigma1,sigma2,a1,a2):
    return lambda x: a1*exp( - (x-dx1)**2 / sigma1**2 ) +  a2*exp( - (x-dx2)**2 / sigma2**2 )
def triple_gaussian(dx1,dx2,dx3,sigma1,sigma2,sigma3,a1,a2,a3):
    return lambda x: abs(a1)*exp( - (x-dx1)**2 / sigma1**2 ) +  abs(a2)*exp( - (x-dx2)**2 / sigma2**2 ) +  abs(a3)*exp( - (x-dx3)**2 / sigma3**2 )
def n_gaussian(dx,sigma,a):
    def g(x):
        v = zeros(len(x))
        for i in range(len(dx)):
            v += a[i] * exp( - ( x - dx[i] )**2 / sigma[i]**2 )
        return v
    return g
def gerr(xarr):
    return lambda p:xarr-gaussian(*p)(*indices(xarr.shape))
def double_gerr(xarr):
    return lambda p:xarr-double_gaussian(*p)(*indices(xarr.shape))
def triple_gerr(xarr):
    return lambda p:xarr-triple_gaussian(*p)(*indices(xarr.shape))
def return_param(xarr,params=None):
    if params == None:
        params = [xarr.argmax(),5,xarr.max()]
    pars, cov, infodict, errmsg, success = optimize.leastsq(gerr(xarr), params, full_output=1)
    return pars
def return_double_param(xarr,params=None):
    if params == None:
        params = [xarr.argmax(),xarr.argmax()+3,4.2,2.3,xarr.max(),xarr.max()/2]
    pars, cov, infodict, errmsg, success = optimize.leastsq(double_gerr(xarr), params, full_output=1)
    return pars
def return_triple_param(xarr,params=None):
    """
    input parameters: center[1-3],width[1-3],amplitude[1-3]
    """
    if params == None:
        params = [xarr.argmax(),xarr.argmax()+3,xarr.argmax(),4.2,2.3,10,xarr.max(),xarr.max()/2.,xarr.max()/5.]
    pars, cov, infodict, errmsg, success = optimize.leastsq(triple_gerr(xarr), params, full_output=1)
    return pars



def adaptive_collapse_gaussfit(cube,axis=2,nsig=3,nrsig=4,prefix='interesting',
        vconv=lambda x: x,xtora=lambda x: x,ytodec=lambda x: x,doplot=True):
    """
    Attempts to fit one or two Gaussians to each spectrum in a data cube and returns the parameters of the fits.
    Adaptively determines where to fit two Gaussian components based on residuals.  Will fit 3 gaussians if a
    two-gaussian fit is not better than a certain threshold (specified by nsig), and those fits will be output
    to images with filename prefix+(coordinate).png.  The 3-gaussian fit parameters will not be returned because
    the automated fitting is very unlikely to get that part right.

    inputs:
    cube - a data cube with two spatial and one spectral dimensions
    axis - the axis of the spectral dimension
    nsig - number of sigma over the mean residual to trigger double-gaussian fitting
           also, cutoff to do any fitting at all
    prefix - the prefix (including directory name) of the output images from 3-gaussian fitting
    doplot - option to turn off plotting of triple-gaussian fits

    vconv,xtora,ytodec - functions to convert the axes from pixel coordinates to ra/dec/velocity coordinates

    returns:
    width_arr1,width_arr2,chi2_arr,offset_arr1,offset_arr2,amp_arr1,amp_arr2
    The Gaussian widths, line centers (in pixel units), amplitudes, and the chi-squared value, not in that order
    These returns are identical to the returns from double_gaussian, but all components will be zero for the second
    gaussian in the case of a single-gaussian fit

    the triple gaussian is guessed to be the double gaussian plus a broad, low-amplitude gaussian.  Ideally this should
    fit outflows reasonably well, but who knows if it really will.
    Another option is to fit a negative-amplitude gaussian to account for self-absorption

    """
    std_coll = cube.std(axis=axis)       # standard deviation of each spectrum
#    mad_coll = MAD(cube,axis=axis)
    mean_std = median(std_coll.ravel())  # median standard deviation (to reject high-signal spectra that have high std)
    if axis > 0:                         # force spectral axis to first axis
        cube = cube.swapaxes(0,axis)
    width_arr = zeros(cube.shape[1:])   # define gaussian param arrays
    width_arr1 = zeros(cube.shape[1:])   # define gaussian param arrays
    width_arr2 = zeros(cube.shape[1:])   # define gaussian param arrays
    amp_arr = zeros(cube.shape[1:])     # define gaussian param arrays
    amp_arr1 = zeros(cube.shape[1:])     # define gaussian param arrays
    amp_arr2 = zeros(cube.shape[1:])     # define gaussian param arrays
    chi2_arr = zeros(cube.shape[1:])     # define gaussian param arrays
    resid_arr = zeros(cube.shape[1:])    # define gaussian param arrays
    offset_arr = zeros(cube.shape[1:])  # define gaussian param arrays 
    offset_arr1 = zeros(cube.shape[1:])  # define gaussian param arrays 
    offset_arr2 = zeros(cube.shape[1:])  # define gaussian param arrays 
    ncarr = (cube.max(axis=0) > mean_std*nsig)        # cutoff: don't fit no-signal spectra
    starttime = time.time()              # timing for output
    print cube.shape
    print "Fitting a total of %i spectra with peak signal above %f" % (ncarr.sum(),mean_std*nsig)
    for i in xrange(cube.shape[1]):      # Loop over all elements for 
        t0 = time.time()
        nspec = (cube[:,i,:].max(axis=0) > mean_std*nsig).sum()
        print "Working on row %d with %d spectra to fit" % (i,nspec) ,
        for j in xrange(cube.shape[2]):
            if cube[:,i,j].max() > mean_std*nsig:
#            if cube[:,i,j].max() > MAD(cube[:,i,j]):            
                pars = return_param(cube[:,i,j])
                width_arr[i,j] = pars[1]
                width_arr1[i,j] = pars[1]
                amp_arr[i,j] = pars[2]
                amp_arr1[i,j] = pars[2]
#                chi2_arr[i,j] = sum(( gerr(cube[:,i,j])(pars) )**2) 
                resid_arr[i,j] = (gerr(cube[:,i,j])(pars)).sum()
                offset_arr[i,j] = pars[0]
                offset_arr1[i,j] = pars[0]
            else: 
                width_arr1[i,j] = numpy.nan
                chi2_arr[i,j] = numpy.nan
                resid_arr[i,j] = numpy.nan
                offset_arr1[i,j] = numpy.nan
        dt = time.time()-t0
        if nspec > 0:
            print "in %f seconds (average: %f)" % (dt,dt/float(nspec))
        else: 
            print 
    chi2_arr = resid_arr**2
    resids = ma.masked_where(numpy.isnan(chi2_arr),chi2_arr) # hide bad values
#    residcut = (resids.mean() + (resids.std() * nrsig) )  # Old versino - used standard deviation and mean
    residcut = (nanmedian(chi2_arr.ravel()) + (MAD(chi2_arr.ravel()) * nrsig) ) # New version: set cutoff by median + nrsig * MAD
    to_refit = (resids > residcut).astype('bool')
#    to_refit[numpy.isnan(to_refit)] = 0
    inds = array(nonzero(to_refit)).transpose()
    dgc,tgc = 0,0
    print "Refitting a total of %i spectra with peak residual above %f" % (to_refit.sum(),residcut)
    f=open("%s_triples.txt" % prefix,'w')
#    vconv = lambda x: (x-p3+1)*dv+v0    # convert to velocity frame
    vind = vconv(arange(cube[:,0,0].shape[0]))
    xind = arange(cube[:,0,0].shape[0])
    for ind in inds:
        i,j = ind
        doublepars = return_double_param(cube[:,i,j])
        old_chi2 = chi2_arr[i,j]
        new_chi2 = sum(square( double_gerr(cube[:,i,j])(doublepars) )) 
        if new_chi2 < old_chi2: # if 2 gaussians is an improvement, use it!
            chi2_arr[i,j] = new_chi2
            width_arr1[i,j] = doublepars[2]
            width_arr2[i,j] = doublepars[3]
            amp_arr1[i,j] = doublepars[4]
            amp_arr2[i,j] = doublepars[5]
            offset_arr1[i,j] = doublepars[0]
            offset_arr2[i,j] = doublepars[1]
            ncarr[i,j] += 1
        if new_chi2 > residcut: # Even if double was better, see if a triple might be better yet [but don't store it in the params arrays!]
            print >>f,"Triple-gaussian fitting at %i,%i (%i'th double, %i'th triple)" % (i,j,dgc,tgc)
            if tgc % 100 == 0:
                print "Triple-gaussian fitting at %i,%i (%i'th double, %i'th triple)" % (i,j,dgc,tgc)
            tgc += 1
            tpguess = [doublepars[0],doublepars[1],(doublepars[0]+doublepars[1])/2.,doublepars[2],doublepars[3],doublepars[2]*5.,doublepars[4],doublepars[5],doublepars[4]/5.]
            triplepars = return_triple_param(cube[:,i,j],params=tpguess)
            pars = [offset_arr[i,j],width_arr[i,j],amp_arr[i,j]]
            if doplot: # if you don't, there's really no point in fitting at all...
                ax = axes([.05,.05,.7,.9])
                plot(vind,cube[:,i,j],color='black',linestyle='steps',linewidth='.5')
                plot(vind,gaussian(*pars)(xind),'r-.',label="Single %f" % ( (gerr(cube[:,i,j])(pars)).sum() ) )
                plot(vind,double_gaussian(*doublepars)(xind),'g--',label="Double %f" % ( (double_gerr(cube[:,i,j])(doublepars)).sum() ))
                plot(vind,triple_gaussian(*triplepars)(xind),'b:',label="Triple %f" % ( (triple_gerr(cube[:,i,j])(triplepars)).sum() ),linewidth=2)
                pars[0] = vconv(pars[0])
                text(1.05,.8,"c1 %3.2f w1 %3.2f a1 %3.2f" % tuple(pars),transform=ax.transAxes,size='smaller')
                dp = [ vconv(doublepars[0]) , doublepars[2], doublepars[4], vconv(doublepars[1]), doublepars[3], doublepars[5] ]
                text(1.05,.6,"c1 %3.2f w1 %3.2f a1 %3.2f\nc2 %3.2f w2 %3.2f a2 %3.2f" % tuple(dp),transform=ax.transAxes,size='smaller')
                tp = [ vconv(triplepars[0]) , triplepars[3], triplepars[6], vconv(triplepars[1]), triplepars[4], triplepars[7], vconv(triplepars[2]), triplepars[5], triplepars[8]  ]
                text(1.05,.4,"c1 %3.2f w1 %3.2f a1 %3.2f\nc2 %3.2f w2 %3.2f a2 %3.2f\nc3 %3.2f w3 %3.2f a3 %3.2f" % tuple(tp),transform=ax.transAxes,size='smaller')
                title("Spectrum at %s %s" % (ratos(xtora(i)),dectos(ytodec(j))) ) 
                legend(loc='best')
                savefig("%s_%s.%s.png" % (prefix,i,j))
                clf()
            ncarr[i,j] += 1
            print >>f,triplepars
        dgc += 1

    f.close()
    print "Total time %f seconds for %i double and %i triple gaussians" % (time.time()-starttime,dgc,tgc)

    return width_arr1,width_arr2,chi2_arr,offset_arr1,offset_arr2,amp_arr1,amp_arr2,ncarr

def collapse_gaussfit(cube,axis=2):
    std_coll = cube.std(axis=axis)
    mean_std = median(std_coll.ravel())
    if axis > 0:
        cube = cube.swapaxes(0,axis)
    width_arr = zeros(cube.shape[1:])
    amp_arr = zeros(cube.shape[1:])
    chi2_arr = zeros(cube.shape[1:])
    offset_arr = zeros(cube.shape[1:])
    starttime = time.time()
    print cube.shape
    print "Fitting a total of %i spectra with peak signal above %f" % ((cube.max(axis=0) > mean_std).sum(),mean_std)
    for i in xrange(cube.shape[1]):
        t0 = time.time()
        nspec = (cube[:,i,:].max(axis=0) > mean_std).sum()
        print "Working on row %d with %d spectra to fit" % (i,nspec) ,
        for j in xrange(cube.shape[2]):
            if cube[:,i,j].max() > mean_std:
                pars = return_param(cube[:,i,j])
                width_arr[i,j] = pars[1]
                chi2_arr[i,j] = sum(( gerr(cube[:,i,j])(pars) )**2) 
                offset_arr[i,j] = pars[0]
                amp_arr[i,j] = pars[2]
            else: 
                width_arr[i,j] = numpy.nan
                chi2_arr[i,j] = numpy.nan
                offset_arr[i,j] = numpy.nan
                amp_arr[i,j] = numpy.nan
        dt = time.time()-t0
        if nspec > 0:
            print "in %f seconds (average: %f)" % (dt,dt/float(nspec))
    print "Total time %f seconds" % (time.time()-starttime)

    return width_arr,offset_arr,amp_arr,chi2_arr

# next step: find 2-gaussian fits
def collapse_double_gaussfit(cube,axis=2):
    std_coll = cube.std(axis=axis)
    mean_std = median(std_coll.ravel())
    if axis > 0:
        cube = cube.swapaxes(0,axis)
    width_arr1 = zeros(cube.shape[1:])
    width_arr2 = zeros(cube.shape[1:])
    amp_arr1 = zeros(cube.shape[1:])
    amp_arr2 = zeros(cube.shape[1:])
    chi2_arr = zeros(cube.shape[1:])
    offset_arr1 = zeros(cube.shape[1:])
    offset_arr2 = zeros(cube.shape[1:])
    starttime = time.time()
    print cube.shape
    print "Fitting a total of %i spectra with peak signal above %f" % ((cube.max(axis=0) > mean_std).sum(),mean_std)
    for i in xrange(cube.shape[1]):
        t0 = time.time()
        nspec = (cube[:,i,:].max(axis=0) > mean_std).sum()
        print "Working on row %d with %d spectra to fit" % (i,nspec) ,
        for j in xrange(cube.shape[2]):
            if cube[:,i,j].max() > mean_std:
                pars = return_double_param(cube[:,i,j])
                width_arr1[i,j] = pars[2]
                width_arr2[i,j] = pars[3]
                amp_arr1[i,j] = pars[4]
                amp_arr2[i,j] = pars[5]
                chi2_arr[i,j] = sum(( double_gerr(cube[:,i,j])(pars) )**2) 
                offset_arr1[i,j] = pars[0]
                offset_arr2[i,j] = pars[1]
            else: 
                width_arr1[i,j] = numpy.nan
                width_arr2[i,j] = numpy.nan
                chi2_arr[i,j] = numpy.nan
                offset_arr1[i,j] = numpy.nan
                offset_arr2[i,j] = numpy.nan
        dt = time.time()-t0
        if nspec > 0:
            print "in %f seconds (average: %f)" % (dt,dt/float(nspec))
    print "Total time %f seconds" % (time.time()-starttime)

    return width_arr1,width_arr2,chi2_arr,offset_arr1,offset_arr2,amp_arr1,amp_arr2

def wrap_collapse_gauss(filename,outprefix,redo='no'):
    """

    redo - if not equal to 'no', then...
    if collapse_gaussfit succeeded (to the extent that the .pysav files were written),
    but some part of the file writing or successive procedures failed, re-do those 
    procedures without redoing the whole collapse
    """
    fitsfile = pyfits.open(filename)
    dv,v0,p3 = fitsfile[0].header['CD3_3'],fitsfile[0].header['CRVAL3'],fitsfile[0].header['CRPIX3']

    cube = fitsfile[0].data
    cube = where(numpy.isnan(cube),0,cube)

    if redo=='no':
        doubleB = asarray(collapse_double_gaussfit(cube,axis=0))
        doubleB[numpy.isnan(doubleB)] = 0
        pickle.dump(doubleB,open('%s_doubleB.pysav' % outprefix,'w'))
    else:
        doubleB = pickle.load(open('%s_doubleB.pysav' % outprefix,'r'))
    db = doubleB
    gcd = double_gaussian(db[3],db[4],db[0],db[1],db[5],db[6])(indices(cube.shape)[0])
    fitsfile[0].data = gcd
    fitsfile.writeto('%s_doublegausscube.fits' % outprefix,clobber=True)

    gcd[numpy.isnan(gcd)] = 0
    doubleResids = cube-gcd
    fitsfile[0].data = doubleResids
    fitsfile.writeto('%s_doublegaussresids.fits' % outprefix,clobber=True)


    #doubleB[4] = (doubleB[4]-v0) / dv + p3-1
    #doubleB[3] = (doubleB[3]-v0) / dv + p3-1
    doubleB[4] = (doubleB[4]-p3+1) * dv + v0
    doubleB[3] = (doubleB[3]-p3+1) * dv + v0
    fitsfile[0].data = asarray(doubleB)
    fitsfile.writeto('%s_doublegausspars.fits' % outprefix,clobber=True)

    if redo=='no':
        singleB = asarray(collapse_gaussfit(cube,axis=0))
        pickle.dump(singleB,open('%s_singleB.pysav' % outprefix,'w'))
    else:
        singleB = pickle.load(open('%s_singleB.pysav' % outprefix,'r'))
    gc = gaussian(singleB[1],singleB[0],singleB[2])(indices(cube.shape)[0])
    singleB[1] = (singleB[1]-p3+1) * dv + v0
    fitsfile[0].data = gc
    fitsfile.writeto('%s_singlegausscube.fits' % outprefix,clobber=True)

    gc[numpy.isnan(gc)]=0
    singleResids = cube-gc
    fitsfile[0].data = singleResids
    fitsfile.writeto('%s_singlegaussresids.fits' % outprefix,clobber=True)
    fitsfile[0].data = asarray(singleB)
    fitsfile.writeto('%s_singlegausspars.fits' % outprefix,clobber=True)

    fitsfile[0].header.__delitem__('CD3_3')
    fitsfile[0].header.__delitem__('CRVAL3')
    fitsfile[0].header.__delitem__('CRPIX3')
    fitsfile[0].header.__delitem__('CUNIT3')
    fitsfile[0].header.__delitem__('CTYPE3')

    doubleResids[numpy.isnan(doubleResids)] = 0
    totalDResids = doubleResids.sum(axis=0)
    fitsfile[0].data = totalDResids
    fitsfile.writeto('%s_doublegauss_totalresids.fits' % outprefix,clobber=True)

    singleResids[numpy.isnan(singleResids)] = 0
    totalSResids = singleResids.sum(axis=0)
    fitsfile[0].data = totalSResids
    fitsfile.writeto('%s_singlegauss_totalresids.fits' % outprefix,clobber=True)

    return singleB,doubleB


def wrap_collapse_adaptive(filename,outprefix,redo='no',nsig=5,nrsig=2,doplot=True):
    """
    redo - if not equal to 'no', then...
    if collapse_gaussfit succeeded (to the extent that the .pysav files were written),
    but some part of the file writing or successive procedures failed, re-do those 
    procedures without redoing the whole collapse
    """
    fitsfile = pyfits.open(filename)
    dv,v0,p3 = fitsfile[0].header['CD3_3'],fitsfile[0].header['CRVAL3'],fitsfile[0].header['CRPIX3']
    dr,r0,p1 = fitsfile[0].header['CD1_1'],fitsfile[0].header['CRVAL1'],fitsfile[0].header['CRPIX1']
    dd,d0,p2 = fitsfile[0].header['CD2_2'],fitsfile[0].header['CRVAL2'],fitsfile[0].header['CRPIX2']
    xtora = lambda x: (x-p1+1)*dr+r0    # convert pixel coordinates to RA/Dec/Velocity
    ytodec = lambda y: (y-p2+1)*dd+d0
    vconv = lambda v: (v-p3+1)*dv+v0

    cube = fitsfile[0].data
    cube = where(numpy.isnan(cube),0,cube)

    if redo=='no':
        adaptB = asarray(adaptive_collapse_gaussfit(cube,axis=0,prefix=outprefix+'_triple',
            nsig=nsig,nrsig=nrsig,vconv=vconv,xtora=xtora,ytodec=ytodec,doplot=doplot))
        adaptB[numpy.isnan(adaptB)] = 0
        pickle.dump(adaptB,open('%s_adaptB.pysav' % outprefix,'w'))
    else:
        adaptB = pickle.load(open('%s_adaptB.pysav' % outprefix,'r'))
    db = adaptB
    gcd = double_gaussian(db[3],db[4],db[0],db[1],db[5],db[6])(indices(cube.shape)[0])
    fitsfile[0].data = gcd
    fitsfile.writeto('%s_adaptgausscube.fits' % outprefix,clobber=True)

    gcd[numpy.isnan(gcd)] = 0
    adaptResids = cube-gcd
    fitsfile[0].data = adaptResids
    fitsfile.writeto('%s_adaptgaussresids.fits' % outprefix,clobber=True)


    #adaptB[4] = (adaptB[4]-v0) / dv + p3-1
    #adaptB[3] = (adaptB[3]-v0) / dv + p3-1
    adaptB[4] = (adaptB[4]-p3+1) * dv + v0
    adaptB[3] = (adaptB[3]-p3+1) * dv + v0
    fitsfile[0].data = asarray(adaptB)
    fitsfile.writeto('%s_adaptgausspars.fits' % outprefix,clobber=True)

    fitsfile[0].header.__delitem__('CD3_3')
    fitsfile[0].header.__delitem__('CRVAL3')
    fitsfile[0].header.__delitem__('CRPIX3')
    fitsfile[0].header.__delitem__('CUNIT3')
    fitsfile[0].header.__delitem__('CTYPE3')

    adaptResids[numpy.isnan(adaptResids)] = 0
    totalDResids = adaptResids.sum(axis=0)
    fitsfile[0].data = totalDResids
    fitsfile.writeto('%s_adaptgauss_totalresids.fits' % outprefix,clobber=True)

    return adaptB




"""
========================================
Generalized redshifted line group fitter
========================================
.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
"""
import numpy as np
import model,fitter

ckms = 2.99792458e5

class redshiftedgroup(object):
    """
    Wrapper for the redshifted group model class.  Specify the offsets and rest
    frequencies when initializing.
    Widths will be matched.
    """

    def __init__(self, freq_dict):
        """
        Initialize the various parameters defining the redshiftedgroup transitions

        freq_dict - frequencies of the indvidual transitions indexed by name
        """
        self.freq_dict = freq_dict

        self.fitter = model.SpectralModel(self,3,
            parnames=['amp','center','width'], 
            parlimited=[(False,False), (False,False), (True,False)], 
            parlimits=[(0,0), (0,0), (0,0)],
            shortvarnames=("A","v","\\sigma"), # specify the parameter names (TeX is OK)
            fitunits='Hz' )

        self.vheight_fitter = model.SpectralModel(fitter.vheightmodel(self),4,
            parnames=['height','amp','center','width'], 
            parlimited=[(False,False), (False,False), (True,False), (False,False), (True,False)], 
            parlimits=[(0,0), (0,0), (0,0), (0,0)],
            shortvarnames=("H","A","v","\\sigma"), # specify the parameter names (TeX is OK)
            fitunits='Hz' )


    def __call__(self, *args, **kwargs):
        """
        Generate a model spectrum given an excitation temperature, optical depth, offset velocity, and velocity width.
        """
        return self.redshiftedgroup(*args,**kwargs)


    def redshiftedgroup(self, xarr, amp=1.0, redshift=0.0, width=1.0, 
            return_components=False ):
        """
        Generate a model spectrum given an amplitude, offset velocity, and velocity width.
        """

        # Convert X-units to frequency in Hz
        xarr = xarr.as_unit('Hz')

        # Generate an optical depth spectrum as a function of the X-axis 
        amp_cumul = np.zeros(len(xarr))
        # Error check: inputing NANs results in meaningless output - return without computing a model
        if np.any(np.isnan((amp,width,redshift))):
            if return_components:
                return [amp_cumul] * len(self.freq_dict)
            else:
                return amp_cumul

        components =[]
        for linename in self.freq_dict:
      
            redline = self.freq_dict[linename] / (1+redshift)
            if width == 0:
                amparr = xarr*0
            else:
                amparr = np.array(amp * np.exp(-(xarr-redline)**2/(2.0*width**2)))
                amparr[amparr!=amparr] = 0 # avoid nans
                components.append( amparr )
            amp_cumul += amparr

        # add a list of the individual 'component' spectra to the total components...

        if return_components:
            return np.array(components)
        else:
            return np.array(amp_cumul)



"""
======================
Simple Gaussian Fitter
======================
"""
import pyspeckit

def fit_gaussians_to_simple_spectra(filename, units='km/s', doplot=True,
        baseline=True, plotresiduals=False, figuresavename=None,
        croprange=None, savename=None, **kwargs):
    """
    As stated in the name title, will fit Gaussians to simple spectra!

    kwargs will be passed to specfit

    *figuresavename* [ None | string ] 
        After fitting, save the figure to this filename if specified

    *croprange* [ list of 2 floats ]
        Crop the spectrum to (min,max) in the specified units

    *savename* [ None | string ]
        After fitting, save the spectrum to this filename

    Note that this wrapper can be used from the command line:

        python fit_gaussians_to_simple_spectra.py spectrum.fits
    
    """

    # load a FITS-compliant spectrum
    spec = pyspeckit.Spectrum(filename)
    if spec.xarr.units != units:
        spec.xarr.frequency_to_velocity()
        spec.xarr.convert_to_unit(units)

    if croprange is not None and len(croprange) == 2:
        spec.crop(*croprange)

    if doplot:
        spec.plotter()
    if baseline:
        spec.baseline()

    spec.specfit(**kwargs)
    spec.specfit(guesses=spec.specfit.modelpars)

    if plotresiduals:
        spec.specfit.plotresiduals()

    if figuresavename is not None:
        spec.plotter.figure.savefig(figuresavename)

    if savename is not None:
        spec.write(savename)

    return spec

if __name__ == "__main__":

    import sys
    import optparse

    parser=optparse.OptionParser()
    parser.add_option("--units","--unit","-u",help="Units.  Default is km/s",default="km/s")
    parser.add_option("--plot","--doplot","-p",help="Plot spectrum.  Default is "+\
            "True, but will not save unless a savename is specified.  You can "+\
            "set the savename with this option.",default=True)
    parser.add_option("--baseline","-b",help="Baseline the spectrum?  Default is True",default=True)
    parser.add_option("--plotresiduals",help="Plot the residual spectrum?  Default is False",default=False)
    parser.add_option("--savename","-s",help="Savename for the .fits or .txt (or other format) file.",default=None)
    parser.add_option("--crop","-c",help="Crop the spectrum?  Syntax is [x1,x2]",default=None)
    parser.set_description("""Fit a gaussian to a spectrum, optionally after baselining.  Save a 
    figure and/or a binary image.  Extra arguments will be passed via kwargs to spectrum.specfit""")

    options,args = parser.parse_args()
    croprange = [int(i) for i in options.crop.strip('[]').split(',')]
    if len(args) > 1: kwargs = args[1:]

    fit_gaussians_to_simple_spectra(args[0],units=options.units,
            doplot=bool(options.plot), baseline=options.baseline,
            plotresiduals=options.plotresiduals, figuresavename=options.plot,
            croprange=croprange, savename=options.savename, **kwargs)



import numpy as np
import arithmetic
import units as units_module
import classes
import pyfits
import headers

def correlate(spectrum1, spectrum2, range=None, units=None, errorweight=False):
    """
    Cross-correlate spectrum1 with spectrum2

    """

    if range is not None:
        spectrum1 = spectrum1.slice(*range, units=units)
        spectrum2 = spectrum2.slice(*range, units=units)

    if not (spectrum1.xarr.shape == spectrum2.xarr.shape) or not all(spectrum1.xarr == spectrum2.xarr):
        spectrum2 = arithmetic.interp(spectrum2, spectrum1)

    data1 = spectrum1.data
    data2 = spectrum2.data

    xcorr = np.correlate(data1, data2, mode='same')

    # very simple propagation of error
    # each element is multiplied, multiplicative error is given such that (sigma_xy/xy)**2 = (sigma_x/x)**2 + (sigma_y/y)**2
    # error = (np.correlate( (spectrum1.error/spectrum1.data)**2 , np.ones(xcorr.shape), mode='same') + 
    #          np.correlate( (spectrum2.error/spectrum2.data)**2 , np.ones(xcorr.shape), mode='same'))**0.5 * xcorr
    # That approach sucks - what if data == 0?
    #
    # this might be more correct: http://arxiv.org/pdf/1006.4069v1.pdf eqn 4
    # but it doesn't quite fit my naive expectations so:
    error = (np.correlate( (spectrum1.error)**2 , np.ones(xcorr.shape), mode='same') + 
             np.correlate( (spectrum2.error)**2 , np.ones(xcorr.shape), mode='same'))**0.5
    
    xarr = spectrum1.xarr
    xrange = xarr.max()-xarr.min()
    xmin = -xrange/2.
    xmax =  xrange/2.
    offset_values = np.linspace(xmin, xmax, len(xarr))

    offset_xarr = units_module.SpectroscopicAxis(offset_values, unit=xarr.units) 

    header = headers.intersection(spectrum1.header, spectrum2.header)
    header.update('CRPIX1',1)
    try:
        header.update('CRVAL1',xmin)
    except ValueError:
        header.update('CRVAL1',xmin.tolist())
    header.update('CDELT1',offset_xarr.cdelt())

    return classes.XCorrSpectrum(xarr=offset_xarr, data=xcorr, header=header, error=error)




"""
====================
Voigt Profile Fitter
====================
"""
import numpy
from numpy.ma import median
from numpy import pi
from pyspeckit.mpfit import mpfit
import matplotlib.cbook as mpcb
from . import fitter

try:
    import scipy.special
    scipyOK = True
except ImportError:
    scipyOK = False
    #print "Voigt profile fitting will fail.  Requires scipy.special"

class voigt_fitter(fitter.SimpleFitter):

    def __init__(self,multisingle='multi'):
        self.npars = 4
        self.npeaks = 1
        if multisingle in ('multi','single'):
            self.multisingle = multisingle
        else:
            raise Exception("multisingle must be multi or single")

    def __call__(self,*args,**kwargs):
        if self.multisingle == 'single':
            return self.onedvoigtfit(*args,**kwargs)
        elif self.multisingle == 'multi':
            return self.multivoigtfit(*args,**kwargs)

    def voigt(self,xarr,amp,xcen,Gfwhm,Lfwhm):
        """
        voigt profile

        V(x,sig,gam) = Re(w(z))/(sig*sqrt(2*pi))
        z = (x+i*gam)/(sig*sqrt(2))

        Converted from 
        http://mail.scipy.org/pipermail/scipy-user/2011-January/028327.html
        """

        if scipyOK:
            tmp = 1.0/scipy.special.wofz(numpy.zeros((len(xarr))) \
                  +1j*numpy.sqrt(numpy.log(2.0))*Lfwhm).real
            tmp = tmp*amp* \
                  scipy.special.wofz(2*numpy.sqrt(numpy.log(2.0))*(xarr-xcen)/Gfwhm+1j* \
                  numpy.sqrt(numpy.log(2.0))*Lfwhm).real
            return tmp
        else:
            raise ImportError("Couldn't import scipy, therefore cannot do voigt profile stuff")

    def n_voigt(self,pars=None):
        """
        Returns a function that sums over N voigt profiles, where N is the length of
        a,dx,Gfwhm,Lfwhm *OR* N = len(pars) / 4

        The background "height" is assumed to be zero (you must "baseline" your
        spectrum before fitting)

        pars  - a list with len(pars) = 3n, assuming a,dx,sigma repeated
        #dx    - offset (velocity center) values
        #width - line widths (Lorentzian FWHM)
        #a     - amplitudes
        """
        if len(pars) % 4 == 0:
            a = [pars[ii] for ii in xrange(0,len(pars),4)]
            dx = [pars[ii] for ii in xrange(1,len(pars),4)]
            Gfwhm = [pars[ii] for ii in xrange(2,len(pars),4)]
            Lfwhm = [pars[ii] for ii in xrange(3,len(pars),4)]
        elif not(len(dx) == len(width) == len(a)):
            raise ValueError("Wrong array lengths! dx: %i  width %i  a: %i" % (len(dx),len(width),len(a)))

        def L(x):
            v = numpy.zeros(len(x))
            for i in range(len(dx)):
                v += self.voigt(x,a[i],dx[i],Gfwhm[i],Lfwhm[i])
            return v
        return L

    def onedvoigtfit(self,xax, data, err=None,
            params=[0,1,0,1,1],fixed=[False,False,False,False,False],
            limitedmin=[False,False,False,True,True],
            limitedmax=[False,False,False,False,False], minpars=[0,0,0,0,0],
            maxpars=[0,0,0,0,0], quiet=True, shh=True, veryverbose=False,
            vheight=True, negamp=False, usemoments=False, **kwargs):
        """
        Inputs:
           xax - x axis
           data - y axis
           err - error corresponding to data

           params - Fit parameters: Height of background, Amplitude, Shift, Width
           fixed - Is parameter fixed?
           limitedmin/minpars - set lower limits on each parameter (default: width>0)
           limitedmax/maxpars - set upper limits on each parameter
           quiet - should MPFIT output each iteration?
           shh - output final parameters?
           usemoments - replace default parameters with moments

        Returns:
           Fit parameters
           Model
           Fit errors
           chi2
        """

        def mpfitfun(x,y,err):
            if err is None:
                def f(p,fjac=None): return [0,(y-self.n_voigt(pars=p[1:])(x)+p[0])]
            else:
                def f(p,fjac=None): return [0,(y-self.n_voigt(pars=p[1:])(x)+p[0])/err]
            return f

        if xax == None:
            xax = numpy.arange(len(data))

        if vheight is False: 
            height = params[0]
            fixed[0] = True
        if usemoments:
            params = moments(xax, data, vheight=vheight, negamp=negamp,
                    veryverbose=veryverbose)
            if vheight is False: params = [height]+params
            if veryverbose: print "OneD moments: h: %g  a: %g  c: %g  w: %g" % tuple(params)

        parinfo = [ {'n':0,'value':params[0],'limits':[minpars[0],maxpars[0]],'limited':[limitedmin[0],limitedmax[0]],'fixed':fixed[0],'parname':"HEIGHT",'error':0} ,
                    {'n':1,'value':params[1],'limits':[minpars[1],maxpars[1]],'limited':[limitedmin[1],limitedmax[1]],'fixed':fixed[1],'parname':"AMPLITUDE",'error':0},
                    {'n':2,'value':params[2],'limits':[minpars[2],maxpars[2]],'limited':[limitedmin[2],limitedmax[2]],'fixed':fixed[2],'parname':"SHIFT",'error':0},
                    {'n':3,'value':params[3],'limits':[minpars[3],maxpars[3]],'limited':[limitedmin[3],limitedmax[3]],'fixed':fixed[3],'parname':"GWIDTH",'error':0},
                    {'n':4,'value':params[4],'limits':[minpars[4],maxpars[4]],'limited':[limitedmin[4],limitedmax[4]],'fixed':fixed[4],'parname':"LWIDTH",'error':0}]

        mp = mpfit(mpfitfun(xax,data,err),parinfo=parinfo,quiet=quiet)
        mpp = mp.params
        if mp.perror is not None: mpperr = mp.perror
        else: mpperr = mpp*0
        chi2 = mp.fnorm

        if mp.status == 0:
            raise Exception(mp.errmsg)

        if (not shh) or veryverbose:
            print "Fit message ",mp.errmsg
            print "Fit status: ",mp.status
            for i,p in enumerate(mpp):
                parinfo[i]['value'] = p
                print parinfo[i]['parname'],p," +/- ",mpperr[i]
            print "Chi2: ",mp.fnorm," Reduced Chi2: ",mp.fnorm/len(data)," DOF:",len(data)-len(mpp)

        self.mp = mp
        self.mpp = mpp[1:]
        self.mpperr = mpperr
        self.model = self.n_voigt(pars=mpp[1:])(xax) 
        return mpp,self.n_voigt(pars=mpp[1:])(xax),mpperr,chi2


    def multivoigtfit(self,xax, data, npeaks=1, err=None, params=[1,0,1,1],
            fixed=[False,False,False,False],
            limitedmin=[False,False,True,True],
            limitedmax=[False,False,False,False], minpars=[0,0,0,0],
            maxpars=[0,0,0,0], quiet=True, shh=True, veryverbose=False,
            **kwargs):
        """
        Fit multiple voigt profiles

        Inputs:
           xax - x axis
           data - y axis
           npeaks - How many voigt profiles to fit?  Default 1 (this could supersede onedgaussfit)
           err - error corresponding to data

         These parameters need to have length = 3*npeaks.  If npeaks > 1 and length = 3, they will
         be replicated npeaks times, otherwise they will be reset to defaults:
           params - Fit parameters: [amplitude, offset, Gfwhm, Lfwhm] * npeaks
                  If len(params) % 4 == 0, npeaks will be set to len(params) / 3
           fixed - Is parameter fixed?
           limitedmin/minpars - set lower limits on each parameter (default: width>0)
           limitedmax/maxpars - set upper limits on each parameter

           quiet - should MPFIT output each iteration?
           shh - output final parameters?

        Returns:
           Fit parameters
           Model
           Fit errors
           chi2
        """

        if len(params) != npeaks and (len(params) / 4) > npeaks:
            npeaks = len(params) / 4 
        self.npeaks = npeaks

        if isinstance(params,numpy.ndarray): params=params.tolist()

        # make sure all various things are the right length; if they're not, fix them using the defaults
        for parlist in (params,fixed,limitedmin,limitedmax,minpars,maxpars):
            if len(parlist) != 4*npeaks:
                # if you leave the defaults, or enter something that can be multiplied by 3 to get to the
                # right number of gaussians, it will just replicate
                if len(parlist) == 4: 
                    parlist *= npeaks 
                elif parlist==params:
                    parlist[:] = [1,0,1,1] * npeaks
                elif parlist==fixed or parlist==limitedmax:
                    parlist[:] = [False,False,False,False] * npeaks
                elif parlist==limitedmin:
                    parlist[:] = [False,False,True,True] * npeaks
                elif parlist==minpars or parlist==maxpars:
                    parlist[:] = [0,0,0,0] * npeaks

        def mpfitfun(x,y,err):
            if err is None:
                def f(p,fjac=None): return [0,(y-self.n_voigt(pars=p)(x))]
            else:
                def f(p,fjac=None): return [0,(y-self.n_voigt(pars=p)(x))/err]
            return f

        if xax == None:
            xax = numpy.arange(len(data))

        parnames = {0:"AMPLITUDE",1:"SHIFT",2:"GFWHM",3:"LFWHM"}

        parinfo = [ {'n':ii, 'value':params[ii],
            'limits':[minpars[ii],maxpars[ii]],
            'limited':[limitedmin[ii],limitedmax[ii]], 'fixed':fixed[ii],
            'parname':parnames[ii%4]+str(ii/4), 'error':ii} 
            for ii in xrange(len(params)) ]

        if veryverbose:
            print "GUESSES: "
            print "\n".join(["%s: %s" % (p['parname'],p['value']) for p in parinfo])

        mp = mpfit(mpfitfun(xax,data,err),parinfo=parinfo,quiet=quiet)
        mpp = mp.params
        if mp.perror is not None: mpperr = mp.perror
        else: mpperr = mpp*0
        chi2 = mp.fnorm

        if mp.status == 0:
            raise Exception(mp.errmsg)

        if not shh:
            print "Fit message ",mp.errmsg
            print "Final fit values: "
            for i,p in enumerate(mpp):
                parinfo[i]['value'] = p
                print parinfo[i]['parname'],p," +/- ",mpperr[i]
            print "Chi2: ",mp.fnorm," Reduced Chi2: ",mp.fnorm/len(data)," DOF:",len(data)-len(mpp)

        self.mp = mp
        self.mpp = mpp
        self.mpperr = mpperr
        self.model = self.n_voigt(pars=mpp)(xax) 
        return mpp,self.n_voigt(pars=mpp)(xax),mpperr,chi2

    def annotations(self):
        label_list = [(
                "$A(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[0+jj*self.npars],self.mpperr[0+jj*self.npars]),
                "$x(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[1+jj*self.npars],self.mpperr[1+jj*self.npars]),
                "$\\sigma_G(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[2+jj*self.npars],self.mpperr[2+jj*self.npars]),
                "$\\sigma_L(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[3+jj*self.npars],self.mpperr[3+jj*self.npars]),
                          ) for jj in range(self.npeaks)]
        labels = tuple(mpcb.flatten(label_list))
        return labels

    def components(self,xarr,modelpars):

        modelcomponents = [ self.voigt(xarr,
            modelpars[4*i],modelpars[4*i+1],modelpars[4*i+2],modelpars[4*i+3])
            for i in range(self.npeaks)]

        return modelcomponents



"""
OH line fitter
"""
import redshiftedgroup


freq_dict={
'OH12':1.61223e9,
'OH11':1.66540e9,
'OH22':1.66736e9,
'OH21':1.72053e9,
}


OH = redshiftedgroup.redshiftedgroup(freq_dict)
OHfitter = OH.fitter
OHvheightfitter = OH.vheight_fitter



""" 
cosmology.py

Author: Jordan Mirocha
Affiliation: University of Colorado at Boulder
Created on 2010-03-01.

Description: Simple cosmology calculator.

Notes: 
    -Everything here uses cgs.
    -I have assumed a flat universe for all calculations, i.e. OmegaCurvatureNow = 0.0.
    -WMAP VII cosmological parameters by default.

"""

import numpy as np
try:
    from scipy.integrate import romberg
    scipyOK=True
except:
    scipyOK=False

c = 29979245800.0
G = 6.673*10**-8
km_per_mpc = 3.08568 * 10**13 * 10**6
cm_per_mpc = 3.08568 * 10**13 * 10**5 * 10**6
sqdeg_per_std = (180.0**2) / (np.pi**2)

class Cosmology:
    def __init__(self, OmegaMatterNow = 0.272, OmegaLambdaNow = 0.728, OmegaBaryonNow = 0.044,
        h_70 = 0.702, sigma_8 = 0.807):
        
        self.OmegaMatterNow = OmegaMatterNow
        self.OmegaLambdaNow = OmegaLambdaNow
        self.OmegaBaryonNow = OmegaBaryonNow
        self.h_70 = h_70
        self.sigma_8 = sigma_8
        self.OmegaCDMNow = self.OmegaMatterNow - self.OmegaBaryonNow
        
        self.HubbleParameterNow = h_70 * 100 / km_per_mpc
        self.CriticalDensityNow = 3 * self.HubbleParameterNow**2 / 8 / np.pi / G

    def ScaleFactor(self, z):
        return 1.0 / (1.0 + z)
        
    def EvolutionFunction(self, z):
        return np.sqrt(self.OmegaMatterNow * (1.0 + z)**3  + self.OmegaLambdaNow)
        
    def HubbleParameter(self, z):	
        return self.HubbleParameterNow * np.sqrt(self.OmegaMatterNow * (1.0 + z)**3 + 
            self.OmegaLambdaNow) 
    
    def OmegaMatter(self, z):
        return self.OmegaMatterNow * (1.0 + z)**3 / self.EvolutionFunction(z)**2
    
    def OmegaLambda(self, z):
	    return self.OmegaLambdaNow / self.EvolutionFunction(z)**2
    
    def MeanMatterDensity(self, z):
        return self.OmegaMatter(z) * self.CriticalDensity(z)
        
    def MeanBaryonDensity(self, z):
        return (self.OmegaBaryonNow / self.OmegaMatterNow) * self.MeanMatterDensity(z)
    
    def CriticalDensity(self, z):
        return (3.0 * self.HubbleParameter(z)**2) / (8.0 * np.pi * G)
        
    def LookbackTime(self, z_i, z_f):
        """
        Returns lookback time from z_i to z_f in seconds, where z_i < z_f.
        """
        Integrand = lambda z: (1.0 / (1.0 + z) / self.EvolutionFunction(z))
        
        if scipyOK:
            return (romberg(Integrand, z_i, z_f) / self.HubbleParameterNow)    
        else: return 0
        
    def TimeToRedshiftConverter(self, z_i, dt):
        """
        Given a redshift and elapsed time dt (in seconds), returns final redshift after dt has passed.
        Uses the high redshift approximation (OmegaLambda -> 0).
        """
        return ((1. + z_i)**(-3. / 2.) + (3. * self.HubbleParameterNow * np.sqrt(self.OmegaMatterNow) * \
            dt / 2.))**(-2. / 3.) - 1.      
    
    def ComovingRadialDistance(self, z_i, z_f):
        """
        Returns comoving radial distance (comoving transverse distance for Omega_K -> 0).
        """
        
        integrand = lambda z: 1. / self.EvolutionFunction(z)
        if scipyOK:
            return romberg(integrand, z_i, z_f) * c / self.HubbleParameterNow / cm_per_mpc
        else: return 0
        
    def LuminosityDistance(self, z_f):
        """
        Returns luminosity distance in Mpc.  Assumes we mean distance from us (z = 0).
        """
        
        return (1. + z_f) * self.ComovingRadialDistance(0., z_f)
        
    



"""
History logger for the spectroscopic toolkit packge

Goal: Save history...
(should this be implemented as a Header class method?)

Author: Adam Ginsburg
Created: 03/18/2011
"""
import time
import pyfits

def write_history(header, string):
    """
    Add a line to the header's history
    """
    if isinstance(header, pyfits.Header):
        hdrstring = time.strftime("%m/%d/%y %H:%M:%S",time.localtime()) + " " + string
        try:
            header.add_history(hdrstring)
        except AttributeError:
            print "WARNING: Error in history writing.  Could not add this string: %s" % hdrstring

    else:
        raise TypeError("Error: header is not a header instance!")



"""
====================
SimpleFitter wrapper 
====================
Adds a variable height (background) component to any model
"""
import numpy
from pyspeckit.mpfit import mpfit
from numpy.ma import median
from pyspeckit.spectrum.moments import moments

class SimpleFitter(object):

    def __init__():
        pass

    def moments(self, *args, **kwargs):
        """
        Get the spectral moments from the moments package
        """
        return moments(*args,**kwargs)


def vheightmodel(zeroheightmodel):
    def vhm(xax, *pars,**kwargs):
        """
        Wrapper function vhm to set variable height.
        Parameter order: height, amplitude, shift, width
        """
        vheight=True
        if 'vheight' in kwargs:
            vheight = kwargs.pop('vheight')
        if vheight:
            return zeroheightmodel(xax, *pars[1:],**kwargs) + pars[0]
        else:
            return zeroheightmodel(xax, *pars[1:],**kwargs)
    vhm.__doc__ += zeroheightmodel.__doc__
    return vhm



"""
Storage for radio spectral line information.
"""

import numpy as np
import copy
import re
from .. import units

try:
    import atpy
    atpyOK = True
except ImportError:
    import readcol
    atpyOK = False

try:
    import query_splatalogue
    webOK = True
except ImportError:
    webOK = False

greekletterlist = ["Alpha","Nu","Beta","Xi","Gamma","Omicron","Delta","Pi","Epsilon","Rho","Zeta","Sigma","Eta","Tau","Theta","Upsilon","Iota","Phi","Kappa","Chi","Lambda","Psi","Mu","Omega"]
greekletterlist += ["alpha","nu","beta","xi","gamma","omicron","delta","pi","epsilon","rho","zeta","sigma","eta","tau","theta","upsilon","iota","phi","kappa","chi","lambda","psi","mu","omega"]
# regular expressions are complicated...
greekletter2  = re.compile("([^\\\]%s)" % ("|[^\\\]".join(greekletterlist)))

greekletter = re.compile("&([A-Za-z][a-z]*);?")
sub = re.compile("<sub>(.*)</sub>")
sup = re.compile("<sup>(.*)</sup>")
junk =  re.compile("&([^a-zA-Z]*);?")
R = re.compile('\W')  # find and remove all non-alphanumeric characters

import os
selfpath = os.path.split(os.path.abspath(__file__))[0]

def LineName(species,qn):
    if greekletter.search(species) is not None:
        #spn = greekletter.sub(greekletter.search(species).groups()[0][0],species)
        name = greekletter.sub(greekletter.search(species).groups()[0][0],qn.replace("(","").replace(")",""))
    else:
        name = species+qn
    if sub.search(name) is not None:
        name = sub.sub("("+sub.search(name).groups()[0]+")",name)
    if sup.search(name) is not None:
        name = sup.sub("("+sup.search(name).groups()[0]+")",name)
    return name.replace(" ","")

def LatexName(species,qn):
    #while greekletter2.search(species) is not None:
    #    species=greekletter2.sub(greekletter2.search(species).groups()[0][0]+"$\\"+greekletter2.search(species).groups()[0][1:]+"$",species)
    if greekletter.search(species) is not None:
        spn = greekletter.sub(r"$\\%s$" % greekletter.search(species).groups()[0],species)
        name = qn.replace("(","$_{").replace(")","}$").replace("&","$\\").replace(";","$")
        if greekletter.search(name) is not None:
            name = greekletter.sub("\\%s" % greekletter.search(name).groups()[0],name) 
    else:
        spn = species
        name = spn+"$(%s)$" % (qn.replace("(","_{").replace(")","}"))
        if greekletter.search(name) is not None:
            name = greekletter.sub(r"\\%s" % greekletter.search(name).groups()[0],name) 
    if junk.search(name) is not None:
        name = junk.sub("",name)
    if sub.search(name) is not None:
        name = sub.sub("_{"+sub.search(name).groups()[0]+"}",name)
    if sup.search(name) is not None:
        name = sup.sub("^{"+sup.search(name).groups()[0]+"}",name)
    if "&" in name:
        raise ValueError("But... that's... not... possible!")
    return name.replace(" ","")

#def radio_lines(minwav=0.0,maxwav=1.0):
#    lines = dict([(LineName(species,qn),[freq,'GHz',True, LatexName(species,qn)]) for
#        species,freq,qn in zip(splat.Species,splat.FreqGHz,splat.ResolvedQNs)])
#
#    #splat = query_splatalogue(minwav,maxwav)


class radio_lines(object):
    def __init__(self, spectrum, voff=0.0, webquery=True, **kwargs):
        """
        Initialize the radio lines class
        Requires a spectrum object 
        """
        self.Spectrum = spectrum

        self.minfreq_GHz = self.Spectrum.xarr.umin(units='GHz')*(1+voff/units.speedoflight_kms)
        self.maxfreq_GHz = self.Spectrum.xarr.umax(units='GHz')*(1+voff/units.speedoflight_kms)

        self.table = get_splat_table(webquery=webquery, 
                minwav=units.speedoflight_ms/self.maxfreq_GHz/1e9,
                maxwav=units.speedoflight_ms/self.minfreq_GHz/1e9,
                waveunits='m')

        self.voff = voff
        self._lines = []
        self._linenames = []

    def show(self, voff=None, ymax_scale=0.8, userecommended=True,
            maxupperstateenergy=None, minupperstateenergy=None, color='r',
            verbose=False, force=False, regexp=None, **kwargs):
        """
        Display vertical lines (using 'vlines') at the position of each
        discovered line
        """
        ymin = self.Spectrum.plotter.ymin
        ymax = self.Spectrum.plotter.ymax
        if ymax < 0:
            ymax_scale = 1.0/ymax_scale

        self.hide()
        voff = self.voff if voff is None else voff

        mask = np.ones(len(self.table),dtype='bool')
        mask *= ((self.table.frequency > self.Spectrum.xarr.umin(units='GHz')*(1+voff/units.speedoflight_kms)) * 
                (self.table.frequency < self.Spectrum.xarr.umax(units='GHz')*(1+voff/units.speedoflight_kms)))

        if userecommended:
            mask *= self.table.frequencyrecommended
        if maxupperstateenergy is not None:
            mask *= (self.table.upperstateenergyK < maxupperstateenergy)
        if minupperstateenergy is not None:
            mask *= (self.table.upperstateenergyK > minupperstateenergy)
        if regexp is not None:
            import re
            reg = re.compile(regexp)
            mask *= np.array([reg.search(sn) is not None for sn in self.table.Species])

        freqoff = voff * 1e3 / units.speedoflight_ms * self.table.frequency[mask]

        if mask.sum() > 0:
            if mask.sum() > 30 and not force:
                print "WARNING: show() will plot %i lines!  Use force=True if you want this to happen anyway." % (mask.sum())
                return
            if verbose: print "Labeled %i lines." % mask.sum()
            self._lines += [self.Spectrum.plotter.axis.vlines( 
                    self.Spectrum.xarr.x_to_coord(self.table.frequency[mask]-freqoff, 'GHz'),
                    ymin, ymax, color=color, **kwargs)]
            self._linenames += [self.Spectrum.plotter.axis.text(
                    self.Spectrum.xarr.x_to_coord(FREQ, 'GHz'),
                    ymax*ymax_scale,
                    NAME,
                    rotation='vertical', color=color)
                for FREQ,NAME in zip(self.table.frequency[mask]-freqoff,self.table.LatexName[mask])]
        else:
            if verbose: print "No lines found in range [%g, %g] GHz" % (self.minfreq_GHz, self.maxfreq_GHz)

        if self.Spectrum.plotter.autorefresh:
            self.Spectrum.plotter.refresh()

    def hide(self):
        """
        Remove all annotations and lines
        """
        for text in self._linenames:
            if text in self.Spectrum.plotter.axis.texts:
                self.Spectrum.plotter.axis.texts.remove(text)
        for text in self._linenames:
            self._linenames.remove(text)
        for line in self._lines:
            if line in self.Spectrum.plotter.axis.collections:
                self.Spectrum.plotter.axis.collections.remove(line)
        for line in self._lines:
            self._lines.remove(line)

        if self.Spectrum.plotter.autorefresh:
            self.Spectrum.plotter.refresh()
 
def get_splat_table(webquery=False, savename=None, **kwargs):
    if webquery and webOK:
        splat = query_splatalogue.query_splatalogue(**kwargs)
        splat.describe()
        splat.rename_column('frequency','FreqGHz')
        splat.FreqGHz /= 1e3
        splat.columns['FreqGHz'].unit = 'GHz'
        splat.rename_column('molecular formula','Species')
        splat.rename_column('quantum numbers','ResolvedQNs')
        try:
            splat.rename_column('chemicalname','ChemicalName')
        except:
            print "Failed to rename chemicalname."
    elif atpyOK:
        splat = atpy.Table(selfpath+"/splatalogue.csv",type='ascii',delimiter=':')
    else:
        splat = readcol.readcol(selfpath+"/splatalogue.csv",fsep=":",asStruct=True)

    for cn in splat.columns:
        if cn != R.sub('',cn):
            splat.rename_column(cn,R.sub('',cn))

    if '' in splat.FreqGHz:
        splat.FreqGHz[splat.FreqGHz == ''] = '-999'
        splat.MeasFreqGHz[splat.MeasFreqGHz == ''] = '-999'
        splat.rename_column('FreqGHz','FreqGHzTxt')
        splat.add_column('FreqGHz',splat.FreqGHzTxt.astype('float'))
        splat.remove_columns('FreqGHzTxt')
        if 'MeasFreqGHz' in splat.columns:
            splat.rename_column('MeasFreqGHz','MeasFreqGHzTxt')
            splat.add_column('MeasFreqGHz',splat.MeasFreqGHzTxt.astype('float'))
            splat.remove_columns('MeasFreqGHzTxt')

    latex_names = [LatexName(species,qn) for species,qn in zip(splat.Species,splat.ResolvedQNs)]
    line_names  = [LineName(species,qn) for species,qn in zip(splat.Species,splat.ResolvedQNs)]
    if hasattr(splat,'add_column'):
        splat.add_column("LineName",line_names)
        splat.add_column("LatexName",latex_names)
        splat.add_column('frequency',splat.FreqGHz)

    if savename is not None:
        splat.write(savename)

    return splat



import numpy as np

def moments(Xax, data, vheight=True, estimator=np.median, negamp=None,
        veryverbose=False, nsigcut=None, noise_guess=None,  **kwargs):
    """Returns (height, amplitude, x, width_x)
    the gaussian parameters of a 1D distribution by calculating its
    moments.  Depending on the input parameters, will only output 
    a subset of the above.
    "height" is the background level
    "amplitude" is the maximum (or minimum) of the data after background subtraction
    "x" is the first moment
    "width_x" is the second moment
    
    If using masked arrays, pass estimator=np.ma.median
    'estimator' is used to measure the background level (height)

    negamp can be used to force the peak negative (True), positive (False),
    or it will be "autodetected" (negamp=None)

    "nsigcut" - try to estimate the noise and only use data above/below nsigma
        above the noise.  Estimate the noise from the data unless passed as a
        keyword

    Theory:
    From first principles (in the absence of noise):
    integral(gaussian) = sqrt(2*pi*sigma^2) * amp
    sigma = integral / amp / sqrt(2*pi)
    
    in the presence of noise, this gets much more complicated

    """

    Xax = np.array(Xax)

    # this is completely absurd.  How the hell do you test for masks?!
    if (data.min() == data.max()):
        return [0]*(3+vheight)
    elif hasattr(data,'mask'):
        # why do I have to do this?  I shouldn't.
        if data.mask.sum() > data.shape[0]-1:
            return [0]*(3+vheight)

    dx = np.abs(np.mean(np.diff(Xax))) # assume a regular grid
    integral = (data*dx).sum()
    height = estimator(data)

    if noise_guess is None:
        noise_guess = data.std() 
    height_cut_low = height-nsigcut*noise_guess if nsigcut is not None else height
    height_cut_high = height+nsigcut*noise_guess if nsigcut is not None else height
    
    # try to figure out whether pos or neg based on the minimum width of the pos/neg peaks
    data_gt_low = data>height_cut_low
    data_lt_low = data<height_cut_low
    Lpeakintegral = integral - height_cut_low*(data_lt_low).sum()*dx - (data[data_gt_low]*dx).sum()
    Lamplitude = data.min()-height
    Lwidth_x = Lpeakintegral / Lamplitude / np.sqrt(2*np.pi)

    data_gt_high = data>height_cut_high
    data_lt_high = data<height_cut_high
    Hpeakintegral = integral - height*(data_gt_high).sum()*dx - (data[data_lt_high]*dx).sum()
    Hamplitude = data.max()-height
    Hwidth_x = Hpeakintegral / Hamplitude / np.sqrt(2*np.pi)

    Lstddev = Xax[data<estimator(data)].std()
    Hstddev = Xax[data>estimator(data)].std()
    #print "Lstddev: %10.3g  Hstddev: %10.3g" % (Lstddev,Hstddev)
    #print "Lwidth_x: %10.3g  Hwidth_x: %10.3g" % (Lwidth_x,Hwidth_x)

    if negamp: # can force the guess to be negative
        xcen,amplitude,width_x = Xax[np.argmin(data)],Lamplitude,Lwidth_x
    elif negamp is None:
        if Hstddev < Lstddev: 
            xcen,amplitude,width_x, = Xax[np.argmax(data)],Hamplitude,Hwidth_x
        else:                                                                   
            xcen,amplitude,width_x, = Xax[np.argmin(data)],Lamplitude,Lwidth_x
    else:  # if negamp==False, make positive
        xcen,amplitude,width_x = Xax[np.argmax(data)],Hamplitude,Hwidth_x

    if veryverbose:
        print "Hstddev: %g   Lstddev: %g" % (Hstddev,Lstddev)
        print "negamp: %s  amp,width,cen Lower: %g, %g   Upper: %g, %g  Center: %g" %\
                (negamp,Lamplitude,Lwidth_x,Hamplitude,Hwidth_x,xcen)
    mylist = [amplitude,xcen,width_x]
    if negamp and amplitude > 0 and veryverbose: 
        print "WARNING: likely fit failure.  negamp=True, but amplitude > 0"
    if negamp is False and amplitude < 0 and veryverbose: 
        print "WARNING: likely fit failure.  negamp=False, but amplitude < 0"
    if np.isnan(width_x) or np.isnan(height) or np.isnan(amplitude):
        raise ValueError("something is nan")
    if vheight:
        mylist = [height] + mylist
    return mylist



mpfit_messages={
-16:"""A parameter or function value has become infinite or an undefined
   number.  This is usually a consequence of numerical overflow in the
   user's model function, which must be avoided.""",
0: "Improper input parameters.",
1: """Both actual and predicted relative reductions in the sum of squares
are at most ftol.""",
2: "Relative error between two consecutive iterates is at most xtol",
3: """Conditions for status = 1 and status = 2 both hold.""",
4: """The cosine of the angle between fvec and any column of the jacobian
  is at most gtol in absolute value.""",
5: """The maximum number of iterations has been reached.""",
6: """ftol is too small. No further reduction in the sum of squares is
  possible.""",
7: """xtol is too small. No further improvement in the approximate solution
  x is possible.""",
8: """gtol is too small. fvec is orthogonal to the columns of the jacobian
  to machine precision."""}

for ii in xrange(-15,0):
    mpfit_messages[ii] = """These are error codes that either MYFUNCT or iterfunct may return to
       terminate the fitting process.  Values from -15 to -1 are reserved
       for the user functions and will not clash with MPFIT."""

from ammonia import ammonia_model,ammonia_model_vtau
# old way from gaussfitter import gaussian_fitter
from inherited_gaussfitter import gaussian_fitter
from inherited_lorentzian import lorentzian_fitter
from inherited_voigtfitter import voigt_fitter
import formaldehyde
from formaldehyde import formaldehyde_fitter,formaldehyde_vheight_fitter
import n2hp
from n2hp import n2hp_vtau_fitter,n2hp_vtau
import hcn
import hyperfine,fitter,model,redshiftedgroup
import hill5infall
import radex_modelgrid
import hydrogen



"""
A simple power-law extinction model

Returns multiplicative factors...
"""
from pyspeckit.spectrum import units

def extinction_function(wavelength, tau0=0, alpha=0, lambda0=2.5, lambda_units='microns'):
    """
    Returns a model with modelfunc replaced...
    """
    if hasattr(wavelength,'as_unit'):
        wavelength = wavelength.as_unit(lambda_units)

    tau_of_lambda = tau0 * ( (wavelength / lambda0)**alpha - 1.0)

    return np.exp(-tau_of_lambda)



"""
====================
HCN Hyperfine Fitter
====================
This is an HCN fitter...
ref for line params: http://www.strw.leidenuniv.nl/~moldata/datafiles/hcn@hfs.dat
"""
import numpy as np
from pyspeckit.mpfit import mpfit
from .. import units
from . import fitter,model,modelgrid
import matplotlib.cbook as mpcb
import copy
try: # for model grid reading
    import pyfits
    pyfitsOK = True
except:
    pyfitsOK = False
try:
    import scipy.interpolate
    import scipy.ndimage
    scipyOK = True
except ImportError:
    scipyOK=False
import hyperfine


freq_dict={
'10-01':88.6339360e9,
'11-01':88.6304160e9,
'12-01':88.6318470e9,
}
aval_dict = {
'10-01':2.4075e-5,
'11-01':2.4075e-5,
'12-01':2.4075e-5,
}
"""
Line strengths of the 15 hyperfine components in J = 1 -  0  transition. The
thickness of the lines indicates their relative weight compared to the others.
Line strengths are normalized in such a way that summing over all initial J = 1
levels gives the degeneracy of the  J = 0 levels, i.e.,  for JF1F = 012,
three for JF1F = 011, and one for JF1F = 010. Thus, the sum over all 15
transitions gives the total spin degeneracy
"""
line_strength_dict = { # effectively the degeneracy per rotation state...
'10-01':1.0,
'11-01':3.0,
'12-01':5.0,
}

relative_strength_total_degeneracy = {
'10-01':9.0,
'11-01':9.0,
'12-01':9.0,
}

line_names = freq_dict.keys()

ckms = units.speedoflight_ms / 1e3 #2.99792458e5
voff_lines_dict = dict([(k,(v-83.6318470e9)/83.6318470e9*ckms) for k,v in freq_dict.iteritems()])

hcn_vtau = hyperfine.hyperfinemodel(line_names, voff_lines_dict, freq_dict, line_strength_dict, relative_strength_total_degeneracy)
hcn_amp = hcn_vtau.ampfitter
hcn_vtau_fitter = hcn_vtau.fitter
hcn_vtau_vheight_fitter = hcn_vtau.vheight_fitter

def hcn_radex(xarr, density=4, column=13, xoff_v=0.0, width=1.0, 
        grid_vwidth=1.0,
        grid_vwidth_scale=False,
        texgrid=None,
        taugrid=None,
        hdr=None,
        path_to_texgrid='',
        path_to_taugrid='',
        temperature_gridnumber=3,
        debug=False,
        verbose=False,
        **kwargs):
    """
    Use a grid of RADEX-computed models to make a model line spectrum

    The RADEX models have to be available somewhere.
    OR they can be passed as arrays.  If as arrays, the form should be:
    texgrid = ((minfreq1,maxfreq1,texgrid1),(minfreq2,maxfreq2,texgrid2))

    xarr must be a SpectroscopicAxis instance
    xoff_v, width are both in km/s

    grid_vwidth is the velocity assumed when computing the grid in km/s
        this is important because tau = modeltau / width (see, e.g., 
        Draine 2011 textbook pgs 219-230)
    grid_vwidth_scale is True or False: False for LVG, True for Sphere
    """

    if texgrid is None and taugrid is None:
        if path_to_texgrid == '' or path_to_taugrid=='':
            raise IOError("Must specify model grids to use.")
        else:
            taugrid = [pyfits.getdata(path_to_taugrid)]
            texgrid = [pyfits.getdata(path_to_texgrid)]
            hdr = pyfits.getheader(path_to_taugrid)
            yinds,xinds = np.indices(taugrid[0].shape[1:])
            densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
            columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
            minfreq = (4.8,)
            maxfreq = (5.0,)
    elif len(taugrid)==len(texgrid) and hdr is not None:
        minfreq,maxfreq,texgrid = zip(*texgrid)
        minfreq,maxfreq,taugrid = zip(*taugrid)
        yinds,xinds = np.indices(taugrid[0].shape[1:])
        densityarr = (xinds+hdr['CRPIX1']-1)*hdr['CD1_1']+hdr['CRVAL1'] # log density
        columnarr  = (yinds+hdr['CRPIX2']-1)*hdr['CD2_2']+hdr['CRVAL2'] # log column
    else:
        raise Exception
    
    # Convert X-units to frequency in GHz
    xarr = copy.copy(xarr)
    xarr.convert_to_unit('Hz', quiet=True)

    tau_nu_cumul = np.zeros(len(xarr))

    gridval1 = np.interp(density, densityarr[0,:], xinds[0,:])
    gridval2 = np.interp(column, columnarr[:,0], yinds[:,0])
    if np.isnan(gridval1) or np.isnan(gridval2):
        raise ValueError("Invalid column/density")

    if scipyOK:
        tau = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[gridval2],[gridval1]]),order=1) for tg in taugrid]
        tex = [scipy.ndimage.map_coordinates(tg[temperature_gridnumber,:,:],np.array([[gridval2],[gridval1]]),order=1) for tg in texgrid]
    else:
        raise ImportError("Couldn't import scipy, therefore cannot interpolate")
    #tau = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,taugrid[temperature_gridnumber,:,:])
    #tex = modelgrid.line_params_2D(gridval1,gridval2,densityarr,columnarr,texgrid[temperature_gridnumber,:,:])

    if verbose:
        print "density %20.12g column %20.12g: tau %20.12g tex %20.12g" % (density, column, tau, tex)

    if debug:
        import pdb; pdb.set_trace()

    return hcn_vtau(xarr,Tex=tex,tau=tau,xoff_v=xoff_v,width=width,**kwargs)




"""
===============
Gaussian fitter
===============

.. moduleauthor:: Adam Ginsburg <adam.g.ginsburg@gmail.com>
Created 3/17/08

Original version available at http://code.google.com/p/agpy/source/browse/trunk/agpy/gaussfitter.py
(the version below uses a Class instead of independent functions)
"""
import numpy
from numpy.ma import median
from numpy import pi
from pyspeckit.mpfit import mpfit
import matplotlib.cbook as mpcb
from . import mpfit_messages
from . import model

class gaussian_fitter(model.SpectralModel):
    """
    A rather complicated Gaussian fitter class.  Inherits from, but overrides
    most components of, :mod:`model.SpectralModel`
    """

    def __init__(self,multisingle='multi'):
        self.npars = 3
        self.npeaks = 1
        self.onepeakgaussfit = self._fourparfitter(self.onepeakgaussian)
        if multisingle in ('multi','single'):
            self.multisingle = multisingle
        else:
            raise Exception("multisingle must be multi or single")

    def __call__(self,*args,**kwargs):
        if self.multisingle == 'single':
            return self.onepeakgaussfit(*args,**kwargs)
        elif self.multisingle == 'multi':
            return self.multigaussfit(*args,**kwargs)

    def onepeakgaussian(self, x,H,A,dx,w):
        """
        Returns a 1-dimensional gaussian of form
        H+A*numpy.exp(-(x-dx)**2/(2*w**2))
        
        [height,amplitude,center,width]
        
        """
        x = numpy.array(x) # make sure xarr is no longer a spectroscopic axis
        return H+A*numpy.exp(-(x-dx)**2/(2*w**2))
        
    def multipeakgaussian(self, x, pars):
        """
        Returns flux at position x due to contributions from multiple Gaussians.
        """
        x = numpy.array(x) # make sure xarr is no longer a spectroscopic axis
        
        pars = numpy.reshape(pars, (len(pars) / 3, 3))
        
        result = 0
        for fit in pars: result += self.onepeakgaussian(x, 0, fit[0], fit[1], fit[2])
        return result
        
    def slope(self, x):
        """
        Return slope at position x for multicomponent Gaussian fit.  Need this in measurements class for
        finding the FWHM of multicomponent lines whose centroids are not identical.
        """    
        
        pars = numpy.reshape(self.mpp, (len(self.mpp) / 3, 3))
        result = 0
        for fit in pars:
            result += self.onepeakgaussian(x, 0, fit[0], fit[1], fit[2]) * (-2. * (x - fit[1]) / 2. / fit[2]**2)
        return result

    def n_gaussian(self, pars=None,a=None,dx=None,sigma=None):
        """
        Returns a function that sums over N gaussians, where N is the length of
        a,dx,sigma *OR* N = len(pars) / 3

        The background "height" is assumed to be zero (you must "baseline" your
        spectrum before fitting)

        pars  - a list with len(pars) = 3n, assuming a,dx,sigma repeated
        dx    - offset (velocity center) values
        sigma - line widths
        a     - amplitudes
        """
        if len(pars) % 3 == 0:
            a = [pars[ii] for ii in xrange(0,len(pars),3)]
            dx = [pars[ii] for ii in xrange(1,len(pars),3)]
            sigma = [pars[ii] for ii in xrange(2,len(pars),3)]
        elif not(len(dx) == len(sigma) == len(a)):
            raise ValueError("Wrong array lengths! dx: %i  sigma: %i  a: %i" % (len(dx),len(sigma),len(a)))

        def g(x):
            v = numpy.zeros(len(x))
            for ii in range(len(pars)/3):
                v += a[ii] * numpy.exp( - ( x - dx[ii] )**2 / (2.0*sigma[ii]**2) )
            return v
        return g

    def multigaussfit(self, xax, data, npeaks=1, err=None, params=[1,0,1],
            fixed=[False,False,False], limitedmin=[False,False,True],
            limitedmax=[False,False,False], minpars=[0,0,0], maxpars=[0,0,0],
            quiet=True, shh=True, veryverbose=False, negamp=None,
            tied = ['', '', ''], parinfo=None, debug=False, **kwargs):
        """
        An improvement on onepeakgaussfit.  Lets you fit multiple gaussians.

        Inputs:
           xax - x axis
           data - y axis
           npeaks - How many gaussians to fit?  Default 1 (this could supersede onepeakgaussfit)
           err - error corresponding to data

         These parameters need to have length = 3*npeaks.  If npeaks > 1 and length = 3, they will
         be replicated npeaks times, otherwise they will be reset to defaults:
           params - Fit parameters: [amplitude, offset, width] * npeaks
                  If len(params) % 3 == 0, npeaks will be set to len(params) / 3
           fixed - Is parameter fixed?
           limitedmin/minpars - set lower limits on each parameter (default: width>0)
           limitedmax/maxpars - set upper limits on each parameter
           tied - link parameters together

           quiet - should MPFIT output each iteration?
           shh - output final parameters?

           kwargs are passed to mpfit

        Returns:
           Fit parameters
           Model
           Fit errors
           chi2
        """
        
        if len(params) != npeaks and (len(params) / 3) > npeaks:
            self.npeaks = len(params) / 3 
        else:
            self.npeaks = npeaks

        if isinstance(params,numpy.ndarray): params=params.tolist()

        # make sure all various things are the right length; if they're not, fix them using the defaults
        # multiformaldehydefit should process negamp directly if kwargs.has_key('negamp') is False: kwargs['negamp'] = None 
        pardict = {"params":params,"fixed":fixed,"limitedmin":limitedmin,"limitedmax":limitedmax,"minpars":minpars,"maxpars":maxpars,"tied":tied}
        for parlistname in pardict:
            parlist = pardict[parlistname]
            if len(parlist) != 3*self.npeaks:
                # if you leave the defaults, or enter something that can be multiplied by 3 to get to the
                # right number of formaldehydeians, it will just replicate
                if veryverbose: print "Correcting length of parameter %s" % parlistname
                if len(parlist) == 3: 
                    parlist *= self.npeaks 
                elif parlistname=="params":
                    parlist[:] = [1,0,1] * self.npeaks
                elif parlistname=="fixed":
                    parlist[:] = [False,False,False] * self.npeaks
                elif parlistname=="limitedmax":
                    if negamp is None: parlist[:] = [False,False,False] * self.npeaks
                    elif negamp is False: parlist[:] = [False,False,False] * self.npeaks
                    else: parlist[:] = [True,False,False] * self.npeaks
                elif parlistname=="limitedmin":
                    if negamp is None: parlist[:] = [False,False,True] * self.npeaks  # Lines can't have negative width!
                    elif negamp is False: parlist[:] = [True,False,True] * self.npeaks
                    else: parlist[:] = [False,False,True] * self.npeaks                   
                elif parlistname=="minpars" or parlistname=="maxpars":
                    parlist[:] = [0,0,0] * self.npeaks
                elif parlistname=="tied":
                    parlist[:] = ['','',''] * self.npeaks
                    
        # mpfit doesn't recognize negamp, so get rid of it now that we're done setting limitedmin/max and min/maxpars
        #if kwargs.has_key('negamp'): kwargs.pop('negamp')

        def mpfitfun(x,y,err):
            if err is None:
                def f(p,fjac=None): return [0,(y-self.n_gaussian(pars=p)(x))]
            else:
                def f(p,fjac=None): return [0,(y-self.n_gaussian(pars=p)(x))/err]
            return f

        if xax is None:
            xax = numpy.arange(len(data))

        parnames = {0:"AMPLITUDE",1:"SHIFT",2:"WIDTH"}

        if parinfo is None:
            parinfo = [ {'n':ii, 'value':params[ii],
                'limits':[minpars[ii],maxpars[ii]],
                'limited':[limitedmin[ii],limitedmax[ii]], 'fixed':fixed[ii],
                'parname':parnames[ii%3]+str(ii/3), 'error':ii, 'tied':tied[ii]} 
                for ii in xrange(len(params)) ]

        if veryverbose:
            print "GUESSES: "
            print "\n".join(["%s: %s" % (p['parname'],p['value']) for p in parinfo])

        if debug: 
            for p in parinfo: print p
            
        mp = mpfit(mpfitfun(xax,data,err),parinfo=parinfo,quiet=quiet,**kwargs)
        mpp = mp.params
        if mp.perror is not None: mpperr = mp.perror
        else: mpperr = mpp*0
        chi2 = mp.fnorm

        if mp.status == 0:
            raise Exception(mp.errmsg)

        if not shh:
            print "Fit status: ",mp.status
            print "Fit error message: ",mp.errmsg
            print "Fit message: ",mpfit_messages[mp.status]
            print "Final fit values: "
            for i,p in enumerate(mpp):
                parinfo[i]['value'] = p
                print parinfo[i]['parname'],p," +/- ",mpperr[i]
            print "Chi2: ",mp.fnorm," Reduced Chi2: ",mp.fnorm/len(data)," DOF:",len(data)-len(mpp)

        self.mp = mp
        self.mpp = mpp
        self.mpperr = mpperr
        self.model = self.n_gaussian(pars=mpp)(xax)
        return mpp,self.n_gaussian(pars=mpp)(xax),mpperr,chi2

    def annotations(self):
        label_list = [(
                "$A(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[0+jj*self.npars],self.mpperr[0+jj*self.npars]),
                "$x(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[1+jj*self.npars],self.mpperr[1+jj*self.npars]),
                "$\\sigma(%i)$=%6.4g $\\pm$ %6.4g" % (jj,self.mpp[2+jj*self.npars],self.mpperr[2+jj*self.npars])
                          ) for jj in range(self.npeaks)]
        labels = tuple(mpcb.flatten(label_list))
        return labels

    def components(self,xarr,modelpars):

        modelcomponents = [ self.onepeakgaussian(xarr,
            0.0,modelpars[3*i],modelpars[3*i+1],modelpars[3*i+2]) for i in range(self.npeaks)]

        return modelcomponents

    def integral(self, modelpars):
        """
        Return the integral of the individual components (ignoring height)
        """
        return self.model.sum()

        # this is the "proper" way to do it, but the above line was used for compatibility with other models
        integ = 0
        if len(modelpars) % 3 == 0:
            for amp,cen,width in numpy.reshape(modelpars,[len(modelpars)/3,3]):
                integ += amp*width*numpy.sqrt(2.0*numpy.pi)

        return integ

    n_modelfunc = n_gaussian




"""
Author: Adam Ginsburg
Created: 3/17/2011
"""
import numpy as np
import numpy.ma as ma
import pyspeckit 
from pyspeckit import spectrum
import operator

def open_3d_fits(filename,wcstype='',average_extra=False, specaxis=3, 
        scale_keyword=None, scale_action=operator.div, **kwargs):
    """
    Grabs all the relevant pieces of a simple FITS-compliant 3d data cube

    Inputs:
        wcstype - the suffix on the WCS type to get to
            velocity/frequency/whatever
        specaxis - Which axis containts the spectrum?  Default 3

    """
    try:
        import astropy.io.fits as pyfits
    except ImportError:
        import pyfits
    if isinstance(filename, pyfits.hdu.image.PrimaryHDU):
        f=[filename]
    elif isinstance(filename, pyfits.hdu.hdulist.HDUList):
        f=filename
    else:
        f = pyfits.open(filename)
    hdr = f[0].header
    cube = ma.array(f[0].data.squeeze())  # remove extra dimensions such as polarization
    cube.mask = np.isnan(cube) + np.isinf(cube)
    if hdr.get('NAXIS') > 3 and average_extra is False:
        print "Extra dimensions in data cube will be ignored (only the first element of that dimension will be kept)."
    while len(cube.shape) > 3:
        if average_extra:
            cube = cube.mean(axis=0)
        else:
            cube = cube[0,...].squeeze()

    # Does CLASS even make cubes?  Maybe....
    if hdr.get('ORIGIN') == 'CLASS-Grenoble':
        # Use the CLASS FITS definition (which is non-standard)
        # http://iram.fr/IRAMFR/GILDAS/doc/html/class-html/node84.html
        # F(n) = RESTFREQ + CRVALi + ( n - CRPIXi ) * CDELTi
        print "Loading a CLASS .fits spectrum"
        dv = -1*hdr.get('CDELT3')
        v0 = hdr.get('RESTFREQ') + hdr.get('CRVAL3')
        p3 = hdr.get('CRPIX3')
    elif hdr.get('CD3_3'+wcstype):
        dv,v0,p3 = hdr['CD3_3'+wcstype],hdr['CRVAL3'+wcstype],hdr['CRPIX3'+wcstype]
    else:
        dv,v0,p3 = hdr['CDELT3'+wcstype],hdr['CRVAL3'+wcstype],hdr['CRPIX3'+wcstype]

    if scale_keyword is not None and hdr.get(scale_keyword):
        print "Found SCALE keyword %s.  Using %s to scale it" % (scale_keyword,scale_action)
        scaleval = hdr.get(scale_keyword)
        cube = scale_action(cube,scaleval)

    xconv = lambda v: ((v-p3+1)*dv+v0)
    xarr = xconv(np.arange(cube.shape[0]))

    XAxis = spectrum.readers.make_axis(xarr,hdr,wcstype=wcstype, specaxis=3, **kwargs)

    return cube,XAxis,hdr,f
